"""Parent class DataN."""
import os
import os.path
from warnings import warn
from typing import Union, NoReturn
from pycifstar import Data, to_data

from cryspy.A_functions_base.function_1_markdown import md_to_html
from cryspy.A_functions_base.function_1_objects import \
    get_functions_of_objet, get_table_html_for_variables

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class DataN(object):
    """Data container of loops and items."""

    def __repr__(self):
        """
        Magic method print() is redefined.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        ls_out = [f"# Object '{self.get_name():}'"]
        for item in self.items:
            if isinstance(item, ItemN):
                ls_out.append(f"{4*' ':}.{item.get_name():}")
            else:
                ls_out.append(f"{4*' ':}.{item.get_name():} (loop)")
                
        method = self.methods_html()
        if method != "":
            ls_out.append(f"\n# Methods:\n{method:}\n")                
        return "\n".join(ls_out)

    def _repr_html_(self):
        """Representation in HTML format."""
        ls_html = [f"<h2>Object '{self.get_name():}'</h2>"]
        ls_html.append(self.attributes_to_html())


        ls_html.append(get_table_html_for_variables(self))

        report = self.report_html()
        if report != "":
            ls_html.append(f"<h2>Description </h2> {report:}")

        ls_html.append(f"<h2>Classes and methods</h2>")
        try:
            names = sorted([obj.__name__ for obj in self.CLASSES_MANDATORY])
            if len(names) != 0:
                ls_html.append("<b>Mandatory classes: </b>")
                ls_html.append(f"{', '.join(names):}.<br>")
        except AttributeError:
            pass
        try:
            names = sorted([obj.__name__ for obj in self.CLASSES_OPTIONAL])
            if len(names) != 0:
                ls_html.append("<b>Optional classes: </b>")
                ls_html.append(f"{', '.join(names):}.<br>")
        except AttributeError:
            pass

        method = self.methods_html()
        if method != "":
            ls_html.append(f"<b>Methods: </b> {method:}")

        return " ".join(ls_html)

    def methods_html(self):
        ls_html = [f".{func_name}" for func_name in
                   get_functions_of_objet(self)]
        return ", ".join(ls_html)+"."
    
    def attributes_to_html(self) -> str:
        """Representation of defined parameters in HTML format.
        """
        ls_html = ["<table>"]
        ls_html.append("<tr><th>Attribute</th><th> Note </th></tr>")
        items_sorted = sorted(self.items, key=lambda item: item.get_name())
        for item in items_sorted:
            item_type = item.__doc__.strip().split("\n")[0]
            ls_html.append(f"<tr><td>.{item.get_name():}</td>\
<td>{item_type:}</td></tr>")
        ls_html.append("</table>")
        return " ".join(ls_html)

    def __str__(self):
        """
        Magic method str() is redefined.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        return self.to_cif()

    def __getattr__(self, name):
        """
        Magic method __getattr__ is slightly changed for special attributes.

        Parameters
        ----------
        name : TYPE
            DESCRIPTION.

        Raises
        ------
        AttributeError
            DESCRIPTION.

        Returns
        -------
        res : TYPE
            DESCRIPTION.

        """
        for item in self.__getattribute__('items'): # self.items gives an error during the parallelization process
            if name.lower() == item.get_name():
                return item
        raise AttributeError(f"Attribute '{name:}' is not defined")

    def is_attribute(self, name):
        """Temporary construction.

        Better to use:

        try:
            obj = self.attribute_name
        except AttributeError as e:
            obj = ...
        """
        for item in self.items:
            if name.lower() == item.get_name():
                return True
        return False

    def __setattr__(self, name, value) -> NoReturn:
        """
        Rules to set attribute.

        Parameters
        ----------
        name : TYPE
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        NoReturn
            DESCRIPTION.

        """
        flag_items, flag_direct = False, True
        if name == "data_name":
            flag_direct = False
            val_new = str(value).strip()
        elif name == "items":
            flag_items = True
            self.add_items(value)
        else:
            cls_value = type(value)
            if cls_value in self.CLASSES:
                l_name = [item.get_name() for item in self.items]
                name_new = value.get_name()
                if name_new in l_name:
                    self.items.pop(l_name.index(name))
                self.items.append(value)
                flag_items, flag_direct = True, False
                if name_new != name:
                    warn(f"Access to variable by '{name_new:}'.", UserWarning)

        if flag_items:
            pass
        elif flag_direct:
            self.__dict__[name] = value
        else:
            self.__dict__[name] = val_new

    def add_items(self, items: list):
        """Add items."""
        l_name = [item.get_name() for item in items]
        
        s_name = set(l_name)

        if len(s_name) != len(l_name):
            warn("Double items were given.", UserWarning)
            items_unique = [items[l_name.index(name)] for name in s_name]
        else:
            items_unique = items
        l_ind_del = []
        for ind_item, item in enumerate(self.items):
            if item.get_name() in s_name:
                l_ind_del.append(ind_item)
        l_ind_del.reverse()
        for ind in l_ind_del:
            self.items.pop(ind)
        for item in items_unique:
            # if isinstance(item, self.CLASSES):
            if isinstance(item, (ItemN, LoopN)):
                self.items.append(item)
            elif type(self) is DataN:
                if issubclass(type(item), (ItemN, LoopN)):
                    self.CLASSES = self.CLASSES + (type(item), )
                    self.CLASSES_OPTIONAL = self.CLASSES_OPTIONAL + (type(item), )
                    self.items.append(item)

    @classmethod
    def make_container(cls, cls_mandatory, cls_optional, prefix):
        """Create DataN object as a container for items."""
        if cls is not DataN:
            warn("The method 'make_container' is used only for DataN class.")
            return
        obj = cls()
        obj.__dict__["CLASSES_MANDATORY"] = cls_mandatory
        obj.__dict__["CLASSES_OPTIONAL"] = cls_optional
        obj.__dict__["CLASSES"] = cls_mandatory+cls_optional
        obj.__dict__["PREFIX"] = prefix
        obj.__dict__["D_DEFAULT"] = {}
        obj.__dict__["items"] = []
        obj.__dict__["data_name"] = ""
        return obj

    @classmethod
    def get_mandatory_attributes(cls, separator: str = "_"):
        """Get a list of mandatory attributes from mandatory classes."""
        l_res = []
        for cls_obj in cls.CLASSES_MANDATORY:
            if issubclass(cls_obj, ItemN):
                cls_item = cls_obj
            else: #LoopN
                cls_item = cls_obj.ITEM_CLASS
            l_res.extend([f"{cls_item.PREFIX:}{separator:}{name_cif:}" 
                              for name_cif in cls_item.ATTR_MANDATORY_CIF])
        return l_res

    def __getitem__(self, name: Union[int, str]):
        """
        Get item by index or predefined index.

        Parameters
        ----------
        name : TYPE
            DESCRIPTION.

        Returns
        -------
        TYPE
            DESCRIPTION.

        """
        if isinstance(name, int):
            return self.items[name]
        elif isinstance(name, str):
            for item in self.items:
                if name.lower() == item.get_name():
                    return item
        return None

    def get_name(self) -> str:
        """Name of object."""
        name = self.PREFIX
        data_name = self.data_name
        if data_name is not None:
            name = f"{name:}_{data_name:}"
        return name.lower()

    def get_variable_names(self) -> list:
        """
        Get names of variable as a list.

        (((#prefix, #NAME), (#prefix, #NAME), (#attribute, #index))

        Returns
        -------
        list
            List of names of variable.

        """
        prefix = self.PREFIX
        data_name = self.data_name
        if isinstance(data_name, str):
            data_name = data_name.lower()
        l_var = []
        for item in self.items:
            l_var.extend(item.get_variable_names())
        l_var_out = [((prefix, data_name), ) + var for var in l_var]
        return l_var_out

    def is_variables(self) -> bool:
        """Define is there variables or not."""
        flag = False
        for item in self.items:
            if item.is_variables():
                flag = True
                break
        return flag

    def get_variable_by_name(self, name: tuple) -> Union[float, int, str]:
        """
        Get variable given by name.

        Parameters
        ----------
        name : tuple
            (((#prefix, #data_name), (#prefix, #loop_name),
               (#attribute, #index_item))

        Returns
        -------
        Union[float, int, str]
            DESCRIPTION.

        """
        prefix = self.PREFIX
        data_name = self.data_name
        if isinstance(data_name, str):
            data_name = data_name.lower()

        prefix_d, prefix_n = name[0], name[1]
        if isinstance(prefix_d[1], str):
            prefix_d_2 = prefix_d[1].lower()
        else:
            prefix_d_2 = prefix_d[1]

        if (prefix_d[0], prefix_d_2) != (prefix, data_name):
            return None

        name_sh = tuple(name[1:])
        for item in self.items:
            if isinstance(item, ItemN):
                prefix = item.PREFIX
            elif isinstance(item, LoopN):
                item_cls = item.ITEM_CLASS
                if item_cls is ItemN:
                    prefix = item[0].PREFIX
                else:
                    prefix = item_cls.PREFIX
            else:
                raise AttributeError(
                    f"Unknown type object '{type(item).__name__:}'")
            if prefix == prefix_n[0]:
                res = item.get_variable_by_name(name_sh)
                if res is not None:
                    return res
        return None

    def set_variable_by_name(self, name: tuple, value) -> NoReturn:
        """
        Set value to variable given by name.

        Parameters
        ----------
        name : tuple
            DESCRIPTION.
        value : TYPE
            DESCRIPTION.

        Returns
        -------
        NoReturn
            DESCRIPTION.

        """
        prefix = self.PREFIX
        data_name = self.data_name

        prefix_d, prefix_n = name[0], name[1]

        if prefix_d != (prefix, data_name):
            return

        name_sh = tuple(name[1:])
        for item in self.items:
            if isinstance(item, ItemN):
                prefix = item.PREFIX
            elif isinstance(item, LoopN):
                item_cls = item.ITEM_CLASS
                if item_cls is ItemN:
                    prefix = item[0].PREFIX
                else:
                    prefix = item_cls.PREFIX
            else:
                raise AttributeError(
                    f"Unknown type object '{type(item).__name__:}'")
            if prefix == prefix_n[0]:
                item.set_variable_by_name(name_sh, value)

    def is_defined(self) -> bool:
        """
        If all mandatory attributes is defined.

        Returns
        -------
        bool
            DESCRIPTION.

        """
        flag = True
        for item in self.items:
            if not(item.is_defined()):
                flag = False
                if isinstance(item, ItemN):
                    warn(f"{item.PREFIX:} is not fully described.",
                         UserWarning)
                    break
                elif isinstance(item, LoopN):
                    warn(f"{item.ITEM_CLASS.PREFIX:} is not fully described.",
                         UserWarning)
                    break
        if flag:
            cls_items = [type(item) for item in self.items]
            for cls_mand in self.CLASSES_MANDATORY:
                if not(cls_mand in cls_items):
                    flag = False
                    warn(f"The object of {cls_mand.__name__:} is not defined.",
                         UserWarning)
                    break
        return flag

    def form_object(self):
        """Form object."""
        pass


    def to_cif(self, separator="_") -> str:
        """Print information about object in string in STAR format.

        Arguments
        ---------
            prefix: prefix in front of label of attribute
            separator: separator between prefix and attribute ("_" or ".")
            flag: for undefined attribute "." will be printed
            flag_minimal if it's True the minimal set of object will be printed

        Returns
        -------
            A string in STAR/CIF format
        """
        ls_out = []
        if self.data_name is None:
            ls_out.append("data_\n")
        else:
            ls_out.append(f"data_{self.data_name:}\n")
        l_item = self.items
        l_s_itemn = [item.to_cif(separator=separator)+"\n"
                     for item in l_item if isinstance(item, ItemN)]
        l_s_loopn = [item.to_cif(separator=separator)+"\n"
                     for item in l_item if isinstance(item, LoopN)]
        if l_s_loopn != []:
            n_max_loop = max([len(_) for _ in l_s_loopn])
            if n_max_loop < 1000:
                n_max_loop = 1000
        else:
            n_max_loop = 10000
        l_n_max_item = [len(_) for _ in l_s_itemn]

        ls_out.extend([_1 for _1, _2 in zip(l_s_itemn, l_n_max_item)
                       if _2 <= n_max_loop])
        ls_out.extend([_ for _ in l_s_loopn])
        ls_out.extend([_1 for _1, _2 in zip(l_s_itemn, l_n_max_item)
                       if _2 > n_max_loop])
        return "\n".join(ls_out)

    @classmethod
    def from_cif(cls, string: str):
        """Generate object from string of CIF format."""
        cif_data = Data()
        flag = cif_data.take_from_string(string)

        cif_items = cif_data.items
        cif_loops = cif_data.loops

        items = []
        flag = True
        n_mandatory = len(cls.CLASSES_MANDATORY)

        # FIXME: if ItemN is not defined in data class it will be loosed. It should be fixed.        
        flag = len(cls.CLASSES) >= n_mandatory
        for cls_ in cls.CLASSES:
            if issubclass(cls_, ItemN):
                prefix_cls = cls_.PREFIX
                if cif_items.is_prefix(prefix_cls):
                    cif_items_prefix = cif_items[prefix_cls]
                    cif_string = str(cif_items_prefix)
                    obj_prefix = cls_.from_cif(cif_string)
                    if obj_prefix is not None:
                        items.append(obj_prefix)
                        flag = True
        
        for cif_loop in cif_loops:
            flag_loop = True
            for i_cls, cls_ in enumerate(cls.CLASSES):
                if issubclass(cls_, LoopN):
                    prefix_cls = cls_.ITEM_CLASS.PREFIX
                    if cif_loop.is_prefix("_"+prefix_cls):
                        cif_string = str(cif_loop)
                        obj_prefix = cls_.from_cif(cif_string)
                        if obj_prefix is not None:
                            items.append(obj_prefix)
                            flag = True
                            flag_loop = False
            if flag_loop:
                loopn = LoopN.from_cif(str(cif_loop))
                if not(loopn is None):
                    items.append(loopn)
                    

        if not(flag):
            return None

        data_name = cif_data.name
        obj = cls(data_name=data_name, items=items)
        obj.form_object()
        return obj

    @classmethod
    def from_cif_file(cls, f_name: str):
        """Read from cif file."""
        if not(os.path.isfile(f_name)):
            raise UserWarning(f"File {f_name:} is not found.")
            return None
        str_from_cif = str(to_data(f_name))
        obj = cls.from_cif(str_from_cif)
        obj.file_input = f_name
        return obj

    def copy(self, data_name: str = ""):
        """Deep copy of object with new data name."""
        s_cif = self.to_cif()
        obj_new = type(self).from_cif(s_cif)
        obj_new.data_name = data_name
        return obj_new

    def report(self):
        return ""

    def report_html(self):
        return md_to_html(self.report())

    def plots(self):
        l_res = []
        for item in self.items:
            for plot in item.plots():
                if plot is not None:
                    l_res.append(plot)
        return l_res

    def fix_variables(self):
        """Fix variables."""
        for item in self.items:
            item.fix_variables()        

    def set_variable(self, name: str, index=None):
        """Set refinement for variable given by name.
        
        Index parameters is used only for objects given as a matrix.
        """
        name_sh = name.strip(".").lower()
        l_name = name_sh.split(".")
        name_1 = l_name[0]
        for item in self.items:
            if name_1 == item.get_name(): 
                if len(l_name) == 1:
                    attr_refs = []
                    if isinstance(item, ItemN):
                        attr_refs = item.ATTR_REF
                    elif isinstance(item, LoopN):
                        item_class = item.ITEM_CLASS
                        if item_class is ItemN:
                            if len(self.items) != 0:
                                attr_refs = item.items[0].ATTR_REF
                        else:
                            attr_refs = item_class.ATTR_REF
                    for attr_ref in attr_refs:
                        item.set_variable(attr_ref, index=index)
                        
                else:
                    item.set_variable(".".join(l_name[1:]), index=index)

    def get_dictionary(self):
        self.form_object()
        res = {}
        res["name"] = getattr(self, "data_name")
        res["type_name"] = self.get_name()
        for item in self.items:
            if isinstance(item, (LoopN, ItemN)):
                dict = item.get_dictionary()
                res.update(dict)
        return res
    
    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        """
        """
        for item in self.items:
            if isinstance(item, (LoopN, ItemN)):
                item.take_parameters_from_dictionary(ddict_diffrn, l_parameter_name=l_parameter_name, l_sigma=l_sigma)
        return None


"""Description of Diffrn class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"

from warnings import warn
from typing import NoReturn, List, Union
import numpy
import scipy
import scipy.misc

from cryspy.A_functions_base.matrix_operations import calc_vector_product_v1_v2_v1

from cryspy.A_functions_base.extinction import calc_extinction_sphere
from cryspy.A_functions_base.flip_ratio import calc_flip_ratio_by_structure_factors
from cryspy.A_functions_base.unit_cell import calc_sthovl_by_unit_cell_parameters, calc_volume_uc_by_unit_cell_parameters, \
    calc_eq_ccs_by_unit_cell_parameters
from cryspy.A_functions_base.structure_factor import calc_f_m_perp_by_sft

from cryspy.A_functions_base.function_1_roots import calc_roots

from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class

from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import \
    DiffrnOrientMatrix
from cryspy.C_item_loop_classes.cl_1_diffrn_refln import DiffrnReflnL
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import Phase
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs

from cryspy.E_data_classes.cl_1_crystal import Crystal
# from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal


class Diffrn(DataN):
    """
    Single diffraction experiment with polarized neutrons.

    Data items in the DIFFRN category record details about
    single diffraction measurements.

    Methods
    -------
        - calc_iint_u_d_flip_ratio
        - calc_fr
        - calc_fm_perp_loc
        - calc_chi_sq
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - estimate_FM


    Attributes
    ----------
        - setup (mandatory)
        - diffrn_radiation (mandatory)
        - diffrn_orient_matrix (mandatory)
        - diffrn_refln (mandatory)
        - extinction
        - phase
        - refln_#phase_name
        - refln_susceptibility_#phase_name
        - refine_ls
    """

    CLASSES_MANDATORY = (Setup, DiffrnOrientMatrix,
                         DiffrnReflnL)
    CLASSES_OPTIONAL = (DiffrnRadiation, Extinction, Phase, ReflnL, ReflnSusceptibilityL,
                        RefineLs, Chi2)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "diffrn"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Diffrn, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def estimate_FM(self, crystal, max_abs_f_mag: float = 5,
                    delta_f_mag: float = 0.001, flag_lambdaover2_correction=False):
        """
        Estimate magnetic structure factor.

        The method calculates magnetic structure factors (only real part) from
        flipping ratios based on give crystal structure.
        The crystal structure should be centrosymmetric
        """
        setup = self.setup
        wavelength = setup.wavelength
        field_norm = setup.field

        diffrn_refln = self.diffrn_refln
        index_h = diffrn_refln.numpy_index_h
        index_k = diffrn_refln.numpy_index_k
        index_l = diffrn_refln.numpy_index_l
        index_hkl = numpy.stack([index_h, index_k, index_l], axis=0)
        fr_exp = diffrn_refln.numpy_fr
        fr_sigma = diffrn_refln.numpy_fr_sigma

        diffrn_radiation = self.diffrn_radiation
        polarization = diffrn_radiation.polarization
        flipper_efficiency = diffrn_radiation.efficiency

        if self.is_attribute("extinction"):
            extinction = self.extinction
            model_extinction = extinction.model
            radius = extinction.radius
            mosaicity = extinction.mosaicity
        else:
            model_extinction = ""
            radius = None
            mosaicity = None

        diffrn_orient_matrix = self.diffrn_orient_matrix
        u_matrix = diffrn_orient_matrix.u
        magnetic_field = field_norm * u_matrix[2, :]

        refln = crystal.calc_refln(index_hkl)
        f_nucl = numpy.array(refln.f_calc, dtype=complex)

        cell = crystal.cell
        unit_cell_parameters = cell.get_unit_cell_parameters()
        sthovl = cell.calc_sthovl(index_hkl[0], index_hkl[1], index_hkl[2])
        eq_ccs, dder_eq_ccs = calc_eq_ccs_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)


        if (setup.is_attribute("ratio_lambdaover2") &
                flag_lambdaover2_correction):
            ratio_lambdaover2 = setup.ratio_lambdaover2
            flag_lambdaover2 = True
        else:
            ratio_lambdaover2 = None
            flag_lambdaover2 = False

        if (flag_lambdaover2 & flag_lambdaover2_correction):
            index_2hkl = 2 * index_hkl
            f_nucl_2hkl = crystal.calc_f_nucl(index_2hkl)
        else:
            f_nucl_2hkl = None

        if (flag_lambdaover2 & flag_lambdaover2_correction):
            sft_2hkl, dder_sft_2hkl = crystal.calc_structure_factor_tensor_ccs(index_2hkl)

            fm_perp_loc_2hkl, dder_f_m_perp = calc_f_m_perp_by_sft(
                sft_2hkl, magnetic_field, eq_ccs,
                flag_sft_ccs=False,
                flag_magnetic_field=False,
                flag_eq_ccs=False)

        else:
            fm_perp_loc_2hkl = None

        def func_extinction(f_sq, flag_f_sq: bool = False):
            model = self.extinction.model
            radius = self.extinction.radius
            mosaicity = self.extinction.mosaicity
            volume_unit_cell = calc_volume_uc_by_unit_cell_parameters(unit_cell_parameters)[0]
            wavelength = self.setup.wavelength
            cos_2theta = 1.-2*numpy.square(sthovl * wavelength)
            flag_radius = True
            flag_mosaicity = True
            flag_volume_unit_cell = True
            flag_cos_2theta = True
            flag_wavelength = True
            return calc_extinction_sphere(
                f_sq, radius, mosaicity, volume_unit_cell, cos_2theta, wavelength,
                model, flag_f_sq=flag_f_sq, flag_radius=flag_radius,
                flag_mosaicity=flag_mosaicity,
                flag_volume_unit_cell=flag_volume_unit_cell,
                flag_cos_2theta=flag_cos_2theta,
                flag_wavelength=flag_wavelength)

        def function_to_find_fm(f_mag):
            """Calc flip ratio.

            Parameters
            ----------
            f_mag : float
                in muB.

            Returns
            -------
            flip_ratio : float
                flip ratio.
            """
            mag_loc_1 = u_matrix[2, 0]*f_mag*0.2695
            mag_loc_2 = u_matrix[2, 1]*f_mag*0.2695
            mag_loc_3 = u_matrix[2, 2]*f_mag*0.2695
            mag_loc = numpy.stack([mag_loc_1, mag_loc_2, mag_loc_3], axis=0)
            f_m_perp, dder_fm_perp_loc = calc_vector_product_v1_v2_v1(eq_ccs, mag_loc, flag_v1=False, flag_v2=False)

            axis_z = u_matrix[2, :]
            f_r, dder_f_r = calc_flip_ratio_by_structure_factors(
                       polarization, flipper_efficiency, f_nucl, f_m_perp, axis_z, func_extinction=func_extinction,
                       c_lambda2=ratio_lambdaover2, f_n_2h=f_nucl_2hkl, f_m_perp_2h=fm_perp_loc_2hkl,
                       flag_polarization=False, flag_flipper=False,
                       flag_f_n=False, flag_f_m_perp=False,
                       flag_c_lambda2=False,
                       flag_f_n_2h=False, flag_f_m_perp_2h=False)

            return f_r

        lFMans, lFMansMin = [], []
        lFMansMin_sigma = []
        for _ind in range(fr_exp.size):
            def f(x):
                return (function_to_find_fm(x)[_ind] - fr_exp[_ind])
            FMans = calc_roots(f, -1*max_abs_f_mag-delta_f_mag, max_abs_f_mag+delta_f_mag)
            lFMans.append(FMans)
            if len(FMans) == 0:
                lFMansMin.append(None)
                lFMansMin_sigma.append(None)
            else:
                FMans_min = min(FMans)
                lFMansMin.append(FMans_min)
                der1 = scipy.misc.derivative(f, FMans_min, 0.1, 1)
                f_mag_sigma = abs(fr_sigma[_ind] / der1)
                lFMansMin_sigma.append(f_mag_sigma)

        ls_out = [f"loop_{crystal.data_name:}"]
        ls_out.append("_estimate_index_h\n_estimate_index_k")
        ls_out.append("_estimate_index_l\n_estimate_F_M\n_estimate_F_M_sigma")
        for ind_h, ind_k, ind_l, f_mag, f_sig in \
                zip(index_h, index_k, index_l, lFMansMin, lFMansMin_sigma):
            ls_out.append(f"{ind_h:} {ind_k:} {ind_l:} {f_mag:}  {f_sig:}")

        res = LoopN.from_cif("\n".join(ls_out))

        return res

    def apply_constraints(self):
        """Apply constraints."""
        pass

    def plots(self):
        l_res = []
        if self.is_attribute("diffrn_refln"):
            diffrn_refln = self.diffrn_refln
            flag_asymmetry = False
            if self.is_attribute("chi2"):
                chi2 = self.chi2
                if chi2.is_attribute("asymmetry"):
                    flag_asymmetry = chi2.asymmetry

            if flag_asymmetry:
                fig_ax = diffrn_refln.plot_asymmetry_vs_asymmetry_calc()
            elif diffrn_refln.is_attribute("fr"):
                fig_ax = diffrn_refln.plot_fr_vs_fr_calc()
            else:
                fig_ax = diffrn_refln.plot_intensity_vs_intensity_calc()

            if fig_ax is not None:
                fig, ax = fig_ax
                ax.set_title(self.data_name + " - "+ax.title.get_text())
                l_res.append((fig, ax))
        return l_res

    def report(self):
        if self.is_attribute("diffrn_refln"):
            diffrn_refln = self.diffrn_refln
            rep_1 = diffrn_refln.report_agreement_factor_exp()
            rep_2 = ""
            if self.is_attribute("diffrn_orient_matrix"):
                ub_matrix = self.diffrn_orient_matrix
                if ub_matrix.is_defined():
                    cell = ub_matrix.cell
                    rep_2 = diffrn_refln.report_chi_sq_exp(cell=cell)
            if rep_2 == "":
                rep_2 = diffrn_refln.report_chi_sq_exp()
            return rep_1 + "\n" + rep_2
        return  ""

    def get_dictionary(self):
        """Form dictionary. See documentation moduel CrysPy using Jupyter notebook.
        """
        ddict = super(Diffrn, self).get_dictionary()

        keys = ddict.keys()
        if (("magnetic_field" in keys) and ("matrix_u" in keys)):
            ddict["magnetic_field"] = ddict["magnetic_field"]*ddict["matrix_u"][6:]

        return ddict


    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        keys = ddict_diffrn.keys()
        if l_parameter_name is not None:
            parameter_label = [hh[0] for hh in l_parameter_name]
        else:
            parameter_label = []
        if "beam_polarization" in keys:
            self.diffrn_radiation.polarization = ddict_diffrn["beam_polarization"]
        if "flipper_efficiency" in keys:
            self.diffrn_radiation.efficiency = ddict_diffrn["flipper_efficiency"]
        if "extinction_mosaicity" in keys:
            self.extinction.mosaicity = ddict_diffrn["extinction_mosaicity"]
        if "extinction_radius" in keys:
            self.extinction.radius = ddict_diffrn["extinction_radius"]
        if "phase_scale" in keys:
            self.phase.scale = ddict_diffrn["phase_scale"]


        if len(parameter_label) > 0:
            for name, sigma in zip(l_parameter_name, l_sigma):
                if name[0] == "wavelength":
                    self.setup.wavelength_sigma = float(sigma)
                if name[0] == "beam_polarization":
                    self.diffrn_radiation.polarization_sigma = float(sigma)
                if name[0] == "flipper_efficiency":
                    self.diffrn_radiation.efficiency_sigma = float(sigma)
                if name[0] == "extinction_mosaicity":
                    self.extinction.mosaicity_sigma = float(sigma)
                if name[0] == "extinction_radius":
                    self.extinction.radius_sigma = float(sigma)
                if name[0] == "phase_scale":
                    self.phase.scale_sigma = float(sigma)

        if (("iint_plus" in keys) and ("iint_minus" in keys)):
            diffrn_refln = self.diffrn_refln
            if diffrn_refln.is_attribute("fr"):
                diffrn_refln.numpy_fr_calc = ddict_diffrn["iint_plus"]/ddict_diffrn["iint_minus"]
            else:
                diffrn_refln.numpy_intensity_calc = ddict_diffrn["intensity_calc"]
            diffrn_refln.numpy_to_items()

        l_dict_diffrn_out_keys = [hh for hh in keys if hh.startswith("dict_in_out_hkl_")]
        for dict_diffrn_out_keys in l_dict_diffrn_out_keys:
            phase_name = dict_diffrn_out_keys[len("dict_in_out_hkl_"):]
            dict_in_out_hkl = ddict_diffrn[dict_diffrn_out_keys]
            dict_in_out_hkl_keys = dict_in_out_hkl.keys()
            l_refln = []
            if (("f_nucl" in dict_in_out_hkl_keys) and ("index_hkl" in dict_in_out_hkl_keys)):
                index_hkl = dict_in_out_hkl["index_hkl"]
                refln = ReflnL(loop_name = phase_name)
                refln.numpy_index_h = index_hkl[0]
                refln.numpy_index_k = index_hkl[1]
                refln.numpy_index_l = index_hkl[2]
                refln.numpy_a_calc = dict_in_out_hkl["f_nucl"].real
                refln.numpy_b_calc = dict_in_out_hkl["f_nucl"].imag
                refln.numpy_sintlambda = dict_in_out_hkl["sthovl"]
                refln.numpy_to_items()
                l_refln.append(refln)
            self.add_items(l_refln)


"""Module define operation with functions 
sym_elems  and magn_sym_elems 

sym_elems: numpy.shape[13, n_elems]
[numerator_x, numerator_y, numerator_z, denominator_xyz,
 r_11, r_12, r_13,
 r_21, r_22, r_23,
 r_31, r_32, r_33]

magn_sym_elems: [22, n_elems]
[numerator_x, numerator_y, numerator_z, denominator_xyz,
 r_11, r_12, r_13,
 r_21, r_22, r_23,
 r_31, r_32, r_33,
 m_11, m_12, m_13,
 m_21, m_22, m_23,
 m_31, m_32, m_33]

Functions
---------
    - form_symm_elems_by_b_i_r_ij
    - calc_numerators_denominator_for_b_i
    - calc_common_denominator
    - calc_rational_sum
    - transform_to_p1

"""
import numpy

# from cryspy.A_functions_base.function_1_strings import 


def form_symm_elems_by_b_i_r_ij(b_i, r_ij):
    """

    Parameters
    ----------
    b_i : fractions
        DESCRIPTION.
    r_ij : int
        DESCRIPTION.

    Returns
    -------
    sym_elems : TYPE
        [b_num_1, b_num_2, b_num_3, r_11, r_12, r_13, r_21, r_22, r_23,
         r_31, r_32, r_33]

    """
    b_num_1, b_num_2, b_num_3, b_den = calc_numerators_denominator_for_b_i(*b_i)
    (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33) = r_ij

    n_1st = 13
    n_2nd = len(r_11)

    sym_elems = numpy.zeros(shape=(n_1st, n_2nd), dtype=int)

    sym_elems[0, :] = b_num_1
    sym_elems[1, :] = b_num_2
    sym_elems[2, :] = b_num_3
    sym_elems[3, :] = b_den

    sym_elems[4, :] = numpy.array(r_11, dtype=int)
    sym_elems[5, :] = numpy.array(r_12, dtype=int)
    sym_elems[6, :] = numpy.array(r_13, dtype=int)
    sym_elems[7, :] = numpy.array(r_21, dtype=int)
    sym_elems[8, :] = numpy.array(r_22, dtype=int)
    sym_elems[9, :] = numpy.array(r_23, dtype=int)
    sym_elems[10, :] = numpy.array(r_31, dtype=int)
    sym_elems[11, :] = numpy.array(r_32, dtype=int)
    sym_elems[12, :] = numpy.array(r_33, dtype=int)

    return sym_elems


def calc_sym_elems_b_with_common_denominator(*argv):
    """Brings elements of several sym_elems_b to a common denominator

    Parameters
    ----------
    *argv : sym_elems_b [4, n_sym_elems]
        

    Returns
    -------
    normalized sym_elems_b [3, n_sym_elems]
        DESCRIPTION.
    denom_common : int
        common denominator.

    """
    l_norm_sym_elems_b, l_norm_denom = [], []
    for sym_elems_b in argv:
        norm_sym_elems_b, norm_denom = calc_normalized_sym_elems_b(sym_elems_b)
        l_norm_sym_elems_b.append(norm_sym_elems_b)
        l_norm_denom.append(norm_denom)
    denom_common = numpy.lcm.reduce(l_norm_denom)
    t_res = tuple([norm_sym_elems_b * (denom_common // norm_denom) 
                   for norm_sym_elems_b, norm_denom
                   in zip(l_norm_sym_elems_b, l_norm_denom)])
    return (*t_res, denom_common)
        

def calc_normalized_sym_elems_b(sym_elems_b):
    """Brings elements of sym_elems_b to a common denominator

    Parameters
    ----------
    sym_elems_b : [{"num_x", "num_y", "num_z", "denominator"}, n_symm_elems]
        DESCRIPTION.

    Returns
    -------
    normalized_sym_elems_b : [{"num_x", "num_y", "num_z"}, n_symm_elems],
    common_denominator

    """
    sym_elems_b_denom = sym_elems_b[3, :]
    denom_common = numpy.lcm.reduce(sym_elems_b_denom)
    coeff = denom_common // sym_elems_b_denom
    norm_sym_elems_b = numpy.stack([sym_elems_b[0, :]*coeff,
                                    sym_elems_b[1, :]*coeff,
                                    sym_elems_b[2, :]*coeff], axis=0)
    return norm_sym_elems_b, denom_common


def calc_numerators_denominator_for_b_i(b_1, b_2, b_3):
    # 6*8 = 48
    den_max =  48

    np_b_1 =  numpy.round(den_max * numpy.array(b_1, dtype=float)).astype(int)
    np_b_2 =  numpy.round(den_max * numpy.array(b_2, dtype=float)).astype(int)
    np_b_3 =  numpy.round(den_max * numpy.array(b_3, dtype=float)).astype(int)
    
    np_den_max = den_max * numpy.ones(shape=np_b_1.shape, dtype=int)
    gcd = numpy.gcd.reduce([np_b_1, np_b_2, np_b_3, np_den_max])

    num_x = np_b_1 // gcd
    num_y = np_b_2 // gcd
    num_z = np_b_3 // gcd
    den = den_max // gcd
    return num_x, num_y, num_z, den


def calc_numerators_and_common_denominator(*argv):
    """Calculate common denominator and corresponding numerators
    
    given parameters:
        numerator_1, denominator_1, numerator_2, denominator_2 and so on

    """
    if len(argv)%2 != 0:
        raise ValueError("Wrong number of input arguments.")

    n_numb = len(argv) // 2
    denoms = [argv[2*ind+1] for ind in range(n_numb)]
    den_common = numpy.lcm.reduce(denoms)

    l_res = []
    for ind in range(n_numb):
        num = argv[2*ind]
        denom = argv[2*ind+1]
        coeff = den_common//denom
        l_res.append(num*coeff)
    return (*l_res, den_common)


def calc_rational_sum(*argv):
    """Calculate sum of rational numbers given by numerator and denominator
    
    given parameters:
        numerator_1, denominator_1, numerator_2, denominator_2 and so on

    """
    denom_nums = calc_numerators_and_common_denominator(*argv)
    den_common = denom_nums[-1]
    sum_num = numpy.sum(denom_nums[:-1], axis=0)
    gcd = numpy.gcd(sum_num, den_common)
    return sum_num // gcd, den_common // gcd


def transform_to_p1(points_abc, sym_elems, shift, centrosymmetry, indexes_xyz,
                    densities):
    """Transoform indexes_xyz and densities to P1 space group.
    

    Parameters
    ----------
    points_abc : numpy.int [6]
        DESCRIPTION.
    sym_elems : numpy.int [13, n_sym_elems]
        DESCRIPTION.
    shift : numpy.int [4, n_shift]
        DESCRIPTION.
    indexes_xyz : numpy.int [3, n_indexes_xyz]
        DESCRIPTION.
    densities : numpy.int [3, n_densities]
        DESCRIPTION.

    Returns
    -------
    indexes_xyz_p1 : numpy.int [3, n_indexes_xyz_p1]
        DESCRIPTION.
    densities_p1 : numpy.int [3, n_densities_p1]
        DESCRIPTION.

    """
    na = numpy.newaxis

    denom_x = points_abc[0]
    denom_y = points_abc[1]
    denom_z = points_abc[2]

    sym_elems_b_num_1 = sym_elems[0]
    sym_elems_b_num_2 = sym_elems[1]
    sym_elems_b_num_3 = sym_elems[2]
    sym_elems_b_denom = sym_elems[3]

    sym_elems_r_11 = sym_elems[4]
    sym_elems_r_12 = sym_elems[5]
    sym_elems_r_13 = sym_elems[6]

    sym_elems_r_21 = sym_elems[7]
    sym_elems_r_22 = sym_elems[8]
    sym_elems_r_23 = sym_elems[9]

    sym_elems_r_31 = sym_elems[10]
    sym_elems_r_32 = sym_elems[11]
    sym_elems_r_33 = sym_elems[12]

    shift_num_1 = shift[0]
    shift_num_2 = shift[1]
    shift_num_3 = shift[2]
    shift_denom = shift[3]

    # checking:
    flag_b_x = numpy.alltrue(numpy.mod(denom_x, sym_elems_b_denom) == 0)
    flag_b_y = numpy.alltrue(numpy.mod(denom_y, sym_elems_b_denom) == 0)
    flag_b_z = numpy.alltrue(numpy.mod(denom_z, sym_elems_b_denom) == 0)
    flag_shift_x = numpy.alltrue(numpy.mod(denom_x, shift_denom) == 0)
    flag_shift_y = numpy.alltrue(numpy.mod(denom_y, shift_denom) == 0)
    flag_shift_z = numpy.alltrue(numpy.mod(denom_z, shift_denom) == 0)

    if not(flag_b_x):
        raise ValueError("The number of points along a is not compatible with \
symmetry operations")
    elif not(flag_b_y):
        raise ValueError("The number of points along b is not compatible with \
symmetry operations")
    elif not(flag_b_z):
        raise ValueError("The number of points along c is not compatible with \
symmetry operations")
    elif not(flag_shift_x):
        raise ValueError("The number of points along a is not compatible with \
shifting parameters")
    elif not(flag_shift_y):
        raise ValueError("The number of points along b is not compatible with \
shifting parameters")
    elif not(flag_shift_z):
        raise ValueError("The number of points along c is not compatible with \
shifting parameters")
    
    
    ind_num_x = indexes_xyz[0]
    ind_num_y = indexes_xyz[1]
    ind_num_z = indexes_xyz[2]
    
    denom_xyz = numpy.lcm.reduce([denom_x, denom_y, denom_z])
    num_x = ind_num_x * denom_xyz // denom_x
    num_y = ind_num_y * denom_xyz // denom_y
    num_z = ind_num_z * denom_xyz // denom_z

    denom_new = sym_elems_b_denom * denom_xyz
    
    num_new_x = ((sym_elems_r_11[na, :]*num_x[:, na] +
                  sym_elems_r_12[na, :]*num_y[:, na] +
                  sym_elems_r_13[na, :]*num_z[:, na]) *
                 sym_elems_b_denom + sym_elems_b_num_1[na, :] *
                 denom_xyz)[:, :, na] + \
        (shift_num_1[na, :]*denom_new[:, na]//shift_denom[na, :])[na, :, :]
        
    num_new_y = ((sym_elems_r_21[na, :]*num_x[:, na] +
                  sym_elems_r_22[na, :]*num_y[:, na] +
                  sym_elems_r_23[na, :]*num_z[:, na]) *
                 sym_elems_b_denom + sym_elems_b_num_2[na, :] *
                 denom_xyz)[:, :, na] + \
        (shift_num_2[na, :]*denom_new[:, na]//shift_denom[na, :])[na, :, :]

    num_new_z = ((sym_elems_r_31[na, :]*num_x[:, na] +
                  sym_elems_r_32[na, :]*num_y[:, na] +
                  sym_elems_r_33[na, :]*num_z[:, na]) *
                 sym_elems_b_denom + sym_elems_b_num_3[na, :] *
                 denom_xyz)[:, :, na] + \
        (shift_num_3[na, :]*denom_new[:, na]//shift_denom[na, :])[na, :, :]

    
    ind_new_x = ((num_new_x * denom_x) // denom_new[na, :, na]) % denom_x
    ind_new_y = ((num_new_y * denom_y) // denom_new[na, :, na]) % denom_y
    ind_new_z = ((num_new_z * denom_z) // denom_new[na, :, na]) % denom_z

    if centrosymmetry:
        ind_new_x = numpy.concatenate((ind_new_x, (-1*ind_new_x) % denom_x),
                                      axis=1)
        ind_new_y = numpy.concatenate((ind_new_y, (-1*ind_new_y) % denom_y),
                                      axis=1)
        ind_new_z = numpy.concatenate((ind_new_z, (-1*ind_new_z) % denom_z),
                                      axis=1)

    
    densities_4d_p1 = densities[:, :, na, na]*numpy.ones(shape=(3, )+
                                                     ind_new_x.shape)
    indexes_xyz_4d_p1 = numpy.stack([ind_new_x, ind_new_y, ind_new_z], axis=0)
    
        
    sh_h = densities_4d_p1.shape
    densities_p1 = densities_4d_p1.reshape((sh_h[0], sh_h[1]*sh_h[2]*sh_h[3]))
    
    sh_h = indexes_xyz_4d_p1.shape
    indexes_xyz_p1 = indexes_xyz_4d_p1.reshape((sh_h[0], sh_h[1]*sh_h[2]*
                                                sh_h[3]))
    
    indexes_xyz_p1_uniq, uniq_ind = numpy.unique(
        indexes_xyz_p1, return_index=True, axis=1)
    densities_p1_uniq = densities_p1[:, uniq_ind]

    return indexes_xyz_p1_uniq, densities_p1_uniq


def get_string_by_symm_elem(symm_elem):
    """Represent symmetry element by string."""
    labels = ("x", "y", "z")
   
    flag_moment = symm_elem.shape[0] > 13
    
    symm_elem_zero = 0*symm_elem
    symm_elem_plus_one = symm_elem_zero + 1
    symm_elem_minus_one = symm_elem_zero - 1

    flag_plus_one = symm_elem_plus_one == symm_elem
    flag_minus_one = symm_elem_minus_one == symm_elem

    str_empty = numpy.zeros(shape=symm_elem_zero.shape, dtype='<3U')

    str_empty[4] = numpy.where(flag_plus_one[4], "+"+labels[0], numpy.where(
        flag_minus_one[4], "-"+labels[0], ""))
    str_empty[5] = numpy.where(flag_plus_one[5], "+"+labels[1], numpy.where(
        flag_minus_one[5], "-"+labels[1], ""))
    str_empty[6] = numpy.where(flag_plus_one[6], "+"+labels[2], numpy.where(
        flag_minus_one[6], "-"+labels[2], ""))

    str_empty[7] = numpy.where(flag_plus_one[7], "+"+labels[0], numpy.where(
        flag_minus_one[7], "-"+labels[0], ""))
    str_empty[8] = numpy.where(flag_plus_one[8], "+"+labels[1], numpy.where(
        flag_minus_one[8], "-"+labels[1], ""))
    str_empty[9] = numpy.where(flag_plus_one[9], "+"+labels[2], numpy.where(
        flag_minus_one[9], "-"+labels[2], ""))

    str_empty[10] = numpy.where(flag_plus_one[10], "+"+labels[0], numpy.where(
        flag_minus_one[10], "-"+labels[0], ""))
    str_empty[11] = numpy.where(flag_plus_one[11], "+"+labels[1], numpy.where(
        flag_minus_one[11], "-"+labels[1], ""))
    str_empty[12] = numpy.where(flag_plus_one[12], "+"+labels[2], numpy.where(
        flag_minus_one[12], "-"+labels[2], ""))

 
    if flag_moment:
        str_empty[13] = numpy.where((flag_plus_one[4]*flag_plus_one[13] | 
                                     flag_minus_one[4]*flag_minus_one[13]),
                                    "+1", "-1")
    char_add = numpy.char.add
    char_strip = numpy.char.strip

    t_x = symm_elem[0]
    t_y = symm_elem[1]
    t_z = symm_elem[2]
    denom = symm_elem[3]

    gcd_x = numpy.gcd(t_x, denom)
    gcd_y = numpy.gcd(t_y, denom)
    gcd_z = numpy.gcd(t_z, denom)

    denom_x, denom_y, denom_z = denom // gcd_x, denom // gcd_y, denom // gcd_z
    t_x_new, t_y_new, t_z_new = t_x // gcd_x, t_y // gcd_y, t_z // gcd_z

    s_t_x_new, s_t_y_new = t_x_new.astype(str), t_y_new.astype(str)
    s_t_z_new = t_z_new.astype(str)

    s_denom_x, s_denom_y = denom_x.astype(str), denom_y.astype(str)
    s_denom_z = denom_z.astype(str)

    s_fract_x = char_add(char_add(s_t_x_new, "/"), s_denom_x)
    s_fract_y = char_add(char_add(s_t_y_new, "/"), s_denom_y)
    s_fract_z = char_add(char_add(s_t_z_new, "/"), s_denom_z)

    s_fract_x[numpy.char.startswith(s_fract_x, "0")] = ""
    s_fract_y[numpy.char.startswith(s_fract_y, "0")] = ""
    s_fract_z[numpy.char.startswith(s_fract_z, "0")] = ""

    flag_x = numpy.char.startswith(s_fract_x, "-")
    s_fract_x = numpy.where(flag_x, s_fract_x, char_add("+", s_fract_x))
    flag_y = numpy.char.startswith(s_fract_y, "-")
    s_fract_y = numpy.where(flag_y, s_fract_y, char_add("+", s_fract_y))
    flag_z = numpy.char.startswith(s_fract_z, "-")
    s_fract_z = numpy.where(flag_z, s_fract_z, char_add("+", s_fract_z))

    s_out_f_1 = char_add(char_add(char_add(str_empty[4], str_empty[5]),
                                  str_empty[6]), s_fract_x)
    s_out_f_2 = char_add(char_add(char_add(str_empty[7], str_empty[8]),
                                  str_empty[9]), s_fract_y)
    s_out_f_3 = char_add(char_add(char_add(str_empty[10], str_empty[11]),
                                  str_empty[12]), s_fract_z)

    s_out_f_1 = char_strip(s_out_f_1, "+")
    s_out_f_2 = char_strip(s_out_f_2, "+")
    s_out_f_3 = char_strip(s_out_f_3, "+")

    s_out_fract = char_add(char_add(char_add(char_add(s_out_f_1, ","),
                                             s_out_f_2), ","), s_out_f_3)
    if flag_moment:
        s_out = char_add(char_add(s_out_fract, ","), str_empty[13])
    else:
        s_out = s_out_fract
    return s_out


import numpy
import scipy
import scipy.special
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class TOFProfile(ItemN):
    """Profile parameters for time-of-flight experiment.

    Attributes
    ----------
        - sigma0, sigma1, sigma2, gamma0, gamma1, gamma2
          alpha0, alpha1, beta0, beta1 (mandatory)
        - size_g, size_l, strain_g, strain_l, peak_shape (optional)

    The variance of the Gaussian component of the peak shape in TOF neutrons is
    given by:

        H_G**2 = (sigma_2+size_g)*d**4 + (sigma_1+strain_g)*d**2 + sigma_0
        H_L = (gamma_2+size_l)*d**2 + (gamma_1+strain_l)*d + gamma_0

    where d is the d-spacing in angstroms.

    Units
    -----
        sigma2, gamma2 are (microsecs/Ang.**2)**2;
        sigma1, gamma1 are (microsecs/Ang.**2);
        sigma0, gamma0 are (microsecs**2);

    """
    ATTR_MANDATORY_NAMES = ("peak_shape", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("peak_shape", )

    ATTR_OPTIONAL_NAMES = (
        "sigma0", "sigma1", "sigma2", "size_g", "strain_g",
        "gamma0", "gamma1", "gamma2", "size_l", "strain_l",
        "alpha0", "alpha1", "alpha2", "beta0", "beta1", "beta00", "beta01", "beta10",
        "r01", "r02", "r03")
    ATTR_OPTIONAL_TYPES = (
        float, float, float, float, float,
        float, float, float, float, float,
        float, float, float, float, float, float, float, float,
        float, float, float)
    ATTR_OPTIONAL_CIF = (
        "sigma0", "sigma1", "sigma2", "size_g", "strain_g",
        "gamma0", "gamma1", "gamma2", "size_l", "strain_l",
        "alpha0", "alpha1", "alpha2", "beta0", "beta1", "beta00", "beta01", "beta10",
        "r01", "r02", "r03")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("sigma2", "sigma1", "sigma0", "gamma2", "gamma1", "gamma0",
                "alpha0", "alpha1", "beta0", "beta1", "size_g", "size_l",
                "strain_g", "strain_l", "beta00", "beta01", "beta10",
                "r01", "r02", "r03", "alpha2")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"sigma0": "{:.5f}", "sigma1": "{:.5f}", "sigma2": "{:.5f}",
                 "gamma0": "{:.5f}", "gamma1": "{:.5f}", "gamma2": "{:.5f}",
                 "alpha0": "{:.5f}", "alpha1": "{:.5f}", "beta0": "{:.5f}",
                 "beta1": "{:.5f}", "size_g": "{:.5f}", "size_l": "{:.5f}",
                 "strain_g": "{:.5f}", "strain_l": "{:.5f}",
                 "beta00": "{:.5f}", "beta01": "{:.5f}", "beta10": "{:.5f}",
                 "r01": "{:.5f}", "r02": "{:.5f}", "r03": "{:.5f}",
                 "alpha2": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {"peak_shape": ["pseudo-Voigt", "Gauss", "type0m"]}

    # default values for the parameters
    D_DEFAULT = {"peak_shape": "pseudo-Voigt"}

    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_profile"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFProfile, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def to_cif(self, separator: str = "_", flag_all_attributes:bool=False) -> str:
        """
        Redefined function to_cif
        """
        d_attributes_peak_shape =  {
            "pseudo-Voigt": [
                "peak_shape",
                "sigma0", "sigma1", "sigma2", "size_g", "strain_g",
                "gamma0", "gamma1", "gamma2", "strain_l", "size_l",
                "alpha0", "alpha1",
                "beta0", "beta1"],
            "Gauss": [
                "peak_shape",
                "sigma0", "sigma1", "sigma2", "size_g", "strain_g",
                "alpha0", "alpha1",
                "beta0", "beta1"],
            "type0m": [
                "peak_shape",
                "sigma0", "sigma1", "sigma2",
                "gamma0", "gamma1", "gamma2",
                "alpha1", "alpha2",
                "beta00", "beta01", "beta10",
                "r01", "r02", "r03"]}
        peak_shape = getattr(self, "peak_shape")
        l_attr_print = d_attributes_peak_shape[peak_shape]
        # REDO: the lines below are fully copied from cl_1_item.py, but the original code should be re-used instead of copying!!!
        ls_out = []
        prefix = self.PREFIX
        for name, name_cif in zip(self.ATTR_NAMES, self.ATTR_CIF):
            if name in l_attr_print:
                if name_cif == "":
                    name_cif = name
                flag_value = self.is_attribute(name)
                if (flag_value | flag_all_attributes):
                    if flag_value:
                        s_val = str(getattr(self, name))
                    else:
                        s_val = "."
                    l_s_val = s_val.split("\n")
                    if len(l_s_val) > 1:
                        ls_out.append(f"_{prefix:}{separator:}{name_cif:}")
                        ls_out.append(";")
                        ls_out.extend(l_s_val)
                        ls_out.append(";")
                    else:
                        if len(s_val.split(" ")) > 1:
                            ls_out.append(
                                f"_{prefix:}{separator:}{name_cif:} \"{s_val:}\"")
                        else:
                            s_val = getattr(self, f"{name:}_as_string")
                            ls_out.append(
                                f"_{prefix:}{separator:}{name_cif:} {s_val:}")
        return "\n".join(ls_out)

    def get_dictionary(self):
        res = {}
        if self.is_attribute("peak_shape"):
            peak_shape = self.peak_shape
        else:
            peak_shape = "pseudo-Voigt"
        res["profile_peak_shape"] = peak_shape
        l_sigma = []
        l_sigma_refinement = []
        for numb in range(3):
            if self.is_attribute(f"sigma{numb:}"):
                l_sigma.append(getattr(self, f"sigma{numb:}"))
                l_sigma_refinement.append(getattr(self, f"sigma{numb:}_refinement"))
            else:
                break
        res["profile_sigmas"] = numpy.array(l_sigma, dtype=float)
        res["flags_profile_sigmas"] = numpy.array(l_sigma_refinement, dtype=float)
        if peak_shape in ["pseudo-Voigt", "type0m"]:
            l_gamma = []
            l_gamma_refinement = []
            for numb in range(3):
                if self.is_attribute(f"gamma{numb:}"):
                    l_gamma.append(getattr(self, f"gamma{numb:}"))
                    l_gamma_refinement.append(getattr(self, f"gamma{numb:}_refinement"))
                else:
                    break
            res["profile_gammas"] = numpy.array(l_gamma, dtype=float)
            res["flags_profile_gammas"] = numpy.array(l_gamma_refinement, dtype=float)
        if peak_shape in ["pseudo-Voigt", "Gauss", ]:
            l_alpha = []
            l_alpha_refinement = []
            for numb in range(2):
                if self.is_attribute(f"alpha{numb:}"):
                    l_alpha.append(getattr(self, f"alpha{numb:}"))
                    l_alpha_refinement.append(getattr(self, f"alpha{numb:}_refinement"))
                else:
                    break
            l_beta = []
            l_beta_refinement = []
            for numb in range(2):
                if self.is_attribute(f"beta{numb:}"):
                    l_beta.append(getattr(self, f"beta{numb:}"))
                    l_beta_refinement.append(getattr(self, f"beta{numb:}_refinement"))
                else:
                    break
            res["profile_alphas"] = numpy.array(l_alpha, dtype=float)
            res["flags_profile_alphas"] = numpy.array(l_alpha_refinement, dtype=float)
            res["profile_betas"] = numpy.array(l_beta, dtype=float)
            res["flags_profile_betas"] = numpy.array(l_beta_refinement, dtype=float)
        elif peak_shape in ["type0m", ]:
            profile_alphas = numpy.array([
                getattr(self, "alpha1"), getattr(self, "alpha2")
                ], dtype=float)
            flags_profile_alphas = numpy.array([
                getattr(self, "alpha1_refinement"), getattr(self, "alpha2_refinement")
                ], dtype=bool)
            profile_betas = numpy.array([
                getattr(self, "beta00"), getattr(self, "beta01"), getattr(self, "beta10")
                ], dtype=float)
            flags_profile_betas = numpy.array([
                getattr(self, "beta00_refinement"),
                getattr(self, "beta01_refinement"),
                getattr(self, "beta10_refinement")
                ], dtype=bool)
            profile_rs = numpy.array([
                getattr(self, "r01"), getattr(self, "r02"), getattr(self, "r03")
                ], dtype=float)
            flags_profile_rs = numpy.array([
                getattr(self, "r01_refinement"),
                getattr(self, "r02_refinement"),
                getattr(self, "r03_refinement")
                ], dtype=bool)
            res["profile_alphas"] = profile_alphas
            res["flags_profile_alphas"] = flags_profile_alphas
            res["profile_betas"] = profile_betas
            res["flags_profile_betas"] = flags_profile_betas
            res["profile_rs"] = profile_rs
            res["flags_profile_rs"] = flags_profile_rs
        return res

class TOFProfileL(LoopN):
    """Profile parameters for time-of-flight experiment.

    """
    ITEM_CLASS = TOFProfile
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFProfileL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name




"""Transformation of density points to P1 space_group

Functions
---------
    -
"""
import numpy

from cryspy.A_functions_base.function_2_mem import transfer_to_density_3d

from cryspy.C_item_loop_classes.cl_1_space_group_symop import \
    SpaceGroupSymopL

from cryspy.C_item_loop_classes.cl_3_density_point import DensityPoint, \
    DensityPointL


def transform_density_point_to_p1(
        density_point : DensityPointL, space_group_symop: SpaceGroupSymopL(),
        points_a: int, points_b: int, points_c: int):
    """Transform density_point to p1.
    
    FIXME: should be deleted
    """
    r_11 = space_group_symop.numpy_r_11.astype(int)
    r_12 = space_group_symop.numpy_r_12.astype(int)
    r_13 = space_group_symop.numpy_r_13.astype(int)

    r_21 = space_group_symop.numpy_r_21.astype(int)
    r_22 = space_group_symop.numpy_r_22.astype(int)
    r_23 = space_group_symop.numpy_r_23.astype(int)

    r_31 = space_group_symop.numpy_r_31.astype(int)
    r_32 = space_group_symop.numpy_r_32.astype(int)
    r_33 = space_group_symop.numpy_r_33.astype(int)

    b_1 = space_group_symop.numpy_b_1.astype(int)
    b_2 = space_group_symop.numpy_b_2.astype(int)
    b_3 = space_group_symop.numpy_b_3.astype(int)

    index_xyz = (numpy.array(density_point.index_x, dtype=int),
                 numpy.array(density_point.index_y, dtype=int),
                 numpy.array(density_point.index_z, dtype=int))

    r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
    b_i = (b_1, b_2, b_3)

    points_abc = (points_a, points_b, points_c)

    den_i = numpy.array(density_point.density, dtype=float)
    # den_ferro_i = numpy.array(density_point.density_ferro, dtype=float)
    # den_aferro_i = numpy.array(density_point.density_antiferro, dtype=float)

    den_3d = transfer_to_density_3d(index_xyz, den_i, points_abc, r_ij, b_i)

    den_1d = den_3d.flatten()
    p_bc = points_b * points_c
    l_item = []
    for i_den, den in enumerate(den_1d):
        ind_x = i_den // p_bc
        h_y = i_den - p_bc*ind_x
        ind_y = h_y // points_c
        ind_z = h_y % points_c
        item = DensityPoint(
            index_x=ind_x, index_y=ind_y, index_z=ind_z,
            fract_x=float(ind_x)/float(points_a),
            fract_y=float(ind_y)/float(points_b),
            fract_z=float(ind_z)/float(points_c),
            density=den)
        l_item.append(item)

    res_obj = DensityPointL()
    res_obj.items = l_item
    return res_obj



"""Description of Pd2d class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"
from warnings import warn
import numpy
from typing import List, NoReturn

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_cos_ang

from cryspy.A_functions_base.matrix_operations import calc_m1_m2_m1t
from cryspy.A_functions_base.unit_cell import calc_matrix_t
from cryspy.A_functions_base.powder_diffraction_const_wavelength import \
    calc_ttheta_phi_by_gamma_nu

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.preocedures import take_items_by_class

from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import PhaseL
from cryspy.C_item_loop_classes.cl_1_pd_peak import PdPeakL
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs
from cryspy.C_item_loop_classes.cl_1_pd2d_background import \
    Pd2dBackground
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_reflex_asymmetry import\
    Pd2dInstrReflexAsymmetry
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_resolution import \
    Pd2dInstrResolution
from cryspy.C_item_loop_classes.cl_1_pd2d_meas import Pd2dMeas
from cryspy.C_item_loop_classes.cl_1_pd2d_proc import Pd2dProc
from cryspy.C_item_loop_classes.cl_1_pd2d_peak import Pd2dPeakL
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_range import Range
from cryspy.C_item_loop_classes.cl_1_texture import TextureL
from cryspy.C_item_loop_classes.cl_1_exclude import ExcludeL

from cryspy.E_data_classes.cl_1_crystal import Crystal

na = numpy.newaxis

class Pd2d(DataN):
    """
    Powder diffraction experiment with polarized or unpolarized neutrons (2d).

    Data items in the DIFFRN category record details about
    2d powder diffraction measurements.

    Methods
    -------
        - calc_iint_u_d_flip_ratio
        - calc_fr
        - calc_fm_perp_loc
        - calc_chi_sq
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - estimate_FM


    Attributes
    ----------
        - setup (mandatory)
        - pd2d_instr_resolution (mandatory)
        - phase (mandatory)
        - pd2d_background (mandatory)
        - pd_meas (mandatory)
        - diffrn_radiation
        - chi2
        - range
        - extinction
        - pd2d_instr_reflex_asymmetry
        - texture
        - exclude
        - pd2d_proc
        - pd2d_peak
        - refine_ls
        - refln_#phase_name
        - refln_susceptibility_#phase_name
    """

    CLASSES_MANDATORY = (Pd2dInstrResolution, PhaseL, DiffrnRadiation,
                         Setup, Range, Pd2dBackground, Pd2dMeas)
    CLASSES_OPTIONAL = (Extinction, ExcludeL, Chi2, Pd2dInstrReflexAsymmetry,
                        TextureL, RefineLs, ReflnL, ReflnSusceptibilityL,
                        Pd2dPeakL, Pd2dProc, PdPeakL)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "pd2d"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Pd2d, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)
    #
    # def calc_profile(self, tth, phi, l_crystal: List[Crystal], l_peak_in=None,
    #                  l_refln_in=None, l_refln_susceptibility_in=None,
    #                  l_dd_in=None, flag_internal: bool = True):
    #     """Calculate intensity for the given diffraction angle."""
    #     proc = Pd2dProc()  # it is output
    #     proc.ttheta = tth
    #     proc.phi = phi
    #
    #     background = self.pd2d_background
    #     int_bkgd = background.interpolate_by_points(tth, phi)
    #     proc.intensity_bkg_calc = int_bkgd
    #
    #     tth_rad = tth*numpy.pi/180.
    #     phi_rad = phi*numpy.pi/180.
    #     cos_theta_1d = numpy.cos(0.5*tth_rad)
    #     sin_phi_1d = numpy.sin(phi_rad)
    #
    #     setup = self.setup
    #     wavelength = setup.wavelength
    #     diffrn_radiation = self.diffrn_radiation
    #     if setup.offset_phi is None:
    #         setup.offset_phi = 0.
    #
    #     phi_0 = setup.offset_phi
    #     phi_rad = (phi-phi_0)*numpy.pi/180.
    #     sin_phi_1d = numpy.sin(phi_rad)
    #
    #     p_u = float(diffrn_radiation.polarization)
    #     p_d = (2.*float(diffrn_radiation.efficiency)-1.)*p_u
    #
    #     tth_min = tth.min()
    #     tth_max = tth.max()+3.
    #     if tth_max > 180.:
    #         tth_max = 180.
    #
    #     sthovl_min = numpy.sin(0.5*tth_min*numpy.pi/180.)/wavelength
    #     sthovl_max = numpy.sin(0.5*tth_max*numpy.pi/180.)/wavelength
    #
    #     res_u_2d = numpy.zeros((tth.shape[0], phi.shape[0]), dtype=float)
    #     res_d_2d = numpy.zeros((tth.shape[0], phi.shape[0]), dtype=float)
    #
    #     try:
    #         texture = self.texture
    #         h_ax, k_ax = float(texture.h_ax[0]), float(texture.k_ax[0]) # FIXME
    #         l_ax = float(texture.l_ax[0])
    #         g_1, g_2 = float(texture.g_1[0]), float(texture.g_2[0])
    #     except AttributeError:
    #         texture = None
    #
    #     phase = self.phase
    #     _h = len(phase.label)
    #     if l_peak_in is None:
    #         l_peak_in = len(phase.label) * [None]
    #     else:
    #         if _h != len(l_peak_in):
    #             l_peak_in = len(phase.label) * [None]
    #     if l_refln_in is None:
    #         l_refln_in = len(phase.label) * [None]
    #     else:
    #         if _h != len(l_refln_in):
    #             l_refln_in = len(phase.label) * [None]
    #     if l_refln_susceptibility_in is None:
    #         l_refln_susceptibility_in = len(phase.label) * [None]
    #     else:
    #         if _h != len(l_refln_susceptibility_in):
    #             l_refln_susceptibility_in = len(phase.label) * [None]
    #     if l_dd_in is None:
    #         l_dd_in = len(phase.label) * [None]
    #     else:
    #         if _h != len(l_dd_in):
    #             l_dd_in = len(phase.label) * [None]
    #
    #     l_peak, l_dd_out = [], []
    #     l_refln, l_refln_s = [], []
    #     for item_phase, peak_in, refln_in, refln_susceptibility_in, dd_in in \
    #             zip(phase.items, l_peak_in, l_refln_in,
    #                 l_refln_susceptibility_in, l_dd_in):
    #         phase_label = item_phase.label
    #         phase_scale = item_phase.scale
    #         try:
    #             phase_igsize = item_phase.igsize
    #             if phase_igsize is None: phase_igsize = 0. # temporary solution
    #         except AttributeError:
    #             phase_igsize = 0.
    #         try:
    #             phase_u = item_phase.u
    #             if phase_u is None: phase_u = 0. # temporary solution
    #         except AttributeError:
    #             phase_u = 0.
    #         try:
    #             phase_v = item_phase.v
    #             if phase_v is None: phase_v = 0. # temporary solution
    #         except AttributeError:
    #             phase_v = 0.
    #         try:
    #             phase_w = item_phase.w
    #             if phase_w is None: phase_w = 0. # temporary solution
    #         except AttributeError:
    #             phase_w = 0.
    #         try:
    #             phase_x = item_phase.x
    #             if phase_x is None: phase_x = 0. # temporary solution
    #         except AttributeError:
    #             phase_x = 0.
    #         try:
    #             phase_y = item_phase.y
    #             if phase_y is None: phase_y = 0. # temporary solution
    #         except AttributeError:
    #             phase_y = 0.
    #
    #         dd_out = {}
    #         for i_crystal, crystal in enumerate(l_crystal):
    #             if crystal.data_name.lower() == phase_label.lower():
    #                 ind_cry = i_crystal
    #                 break
    #         if ind_cry is None:
    #             warn(f"Crystal with name '{phase_label:}' is not found.",
    #                  UserWarning)
    #             return
    #
    #         crystal = l_crystal[ind_cry]
    #
    #         cell = crystal.cell
    #         # space_group = crystal.space_group
    #
    #         if peak_in is not None:
    #             index_h = peak_in.numpy_index_h
    #             index_k = peak_in.numpy_index_k
    #             index_l = peak_in.numpy_index_l
    #             mult = peak_in.numpy_index_mult
    #         else:
    #             if texture is None:
    #                 ind_hkl_mult = crystal.calc_hkl(sthovl_min, sthovl_max)
    #             else:
    #                 ind_hkl_mult = crystal.calc_hkl_in_range(sthovl_min, sthovl_max)
    #             index_h, index_k, index_l, mult = ind_hkl_mult[0], ind_hkl_mult[1], ind_hkl_mult[2], ind_hkl_mult[3]
    #
    #         peak = Pd2dPeakL(loop_name=phase_label)
    #         peak.numpy_index_h = index_h
    #         peak.numpy_index_k = index_k
    #         peak.numpy_index_l = index_l
    #         peak.numpy_index_mult = mult
    #         peak.numpy_to_items()
    #
    #         cond_1 = len(crystal.get_variable_names()) == 0
    #         cond_2 = (peak_in is not None) & (refln_in is not None)
    #
    #         if (cond_1 & cond_2):
    #             f_nucl_sq = peak_in.numpy_f_nucl_sq
    #             f_m_p_sin_sq = peak_in.numpy_f_m_p_sin_sq
    #             f_m_p_cos_sq = peak_in.numpy_f_m_p_cos_sq
    #             cross_sin = peak_in.numpy_cross_sin
    #             refln = refln_in
    #             refln_s = refln_susceptibility_in
    #         else:
    #             f_nucl_sq, f_m_p_sin_sq, f_m_p_cos_sq, cross_sin, refln, \
    #                 refln_s = self.calc_for_iint(index_h, index_k, index_l,
    #                                              crystal)
    #             refln.loop_name = phase_label
    #             refln_s.loop_name = phase_label
    #         l_refln.append(refln)
    #         l_refln_s.append(refln_s)
    #
    #         peak.numpy_f_nucl_sq = f_nucl_sq
    #         peak.numpy_f_m_p_sin_sq = f_m_p_sin_sq
    #         peak.numpy_f_m_p_cos_sq = f_m_p_cos_sq
    #         peak.numpy_cross_sin = cross_sin
    #
    #         cond_1 = dd_in is not None
    #         cond_2 = ((len(crystal.get_variable_names()) == 0) &
    #                   (not(diffrn_radiation.polarization_refinement)) &
    #                   (not(diffrn_radiation.efficiency_refinement)))
    #         if cond_1 & cond_2:
    #             iint_u_3d, iint_d_3d = dd_in["iint_u_3d"], dd_in["iint_d_3d"]
    #             cos_theta_3d = dd_in["cos_theta_3d"]
    #             sin_phi_3d = dd_in["sin_phi_3d"]
    #         else:
    #             cos_theta_3d, sin_phi_3d, mult_f_n_3d = numpy.meshgrid(
    #                 cos_theta_1d, sin_phi_1d, mult*f_nucl_sq, indexing="ij")
    #
    #             mult_f_m_c_3d = numpy.meshgrid(
    #                 tth_rad, phi_rad, mult*f_m_p_cos_sq, indexing="ij")[2]
    #
    #             hh_u_s_3d = numpy.meshgrid(
    #                 tth_rad, phi_rad, mult*(f_m_p_sin_sq+p_u*cross_sin),
    #                 indexing="ij")[2]
    #             hh_d_s_3d = numpy.meshgrid(
    #                 tth_rad, phi_rad, mult*(f_m_p_sin_sq-p_d*cross_sin),
    #                 indexing="ij")[2]
    #
    #             c_a_sq_3d = (cos_theta_3d * sin_phi_3d)**2
    #             s_a_sq_3d = 1.-c_a_sq_3d
    #             iint_u_3d = (mult_f_n_3d + hh_u_s_3d*s_a_sq_3d +
    #                          mult_f_m_c_3d*c_a_sq_3d)
    #             iint_d_3d = (mult_f_n_3d + hh_d_s_3d*s_a_sq_3d +
    #                          mult_f_m_c_3d*c_a_sq_3d)
    #         dd_out["cos_theta_3d"] = cos_theta_3d
    #         dd_out["sin_phi_3d"] = sin_phi_3d
    #         dd_out["iint_u_3d"] = iint_u_3d
    #         dd_out["iint_d_3d"] = iint_d_3d
    #
    #         sthovl_hkl = cell.calc_sthovl(index_h, index_k, index_l)
    #         tth_hkl_rad = numpy.where(sthovl_hkl*wavelength < 1.,
    #                                   2.*numpy.arcsin(sthovl_hkl*wavelength),
    #                                   numpy.pi)
    #         tth_hkl = tth_hkl_rad*180./numpy.pi
    #
    #         cond_1 = dd_in is not None
    #         flag_phase_uvwxy_ref = (
    #             item_phase.igsize_refinement | item_phase.u_refinement |
    #             item_phase.v_refinement | item_phase.w_refinement |
    #             item_phase.x_refinement | item_phase.y_refinement)
    #         cond_2 = ((not(flag_phase_uvwxy_ref)) &
    #                   (not(self.pd2d_instr_resolution.is_variables())) &
    #                   (not(setup.is_variables())))
    #         if cond_1 & cond_2:
    #             profile_3d, tth_zs = dd_in["profile_3d"], dd_in["tth_zs"]
    #             h_pv = dd_in["h_pv"]
    #         else:
    #             profile_3d, tth_zs, h_pv = self.calc_shape_profile(
    #                 tth, phi, tth_hkl, phase_igsize=phase_igsize,
    #                 phase_u=phase_u, phase_v=phase_v, phase_w=phase_w,
    #                 phase_x=phase_x, phase_y=phase_y)
    #         dd_out["profile_3d"] = profile_3d
    #         dd_out["tth_zs"] = tth_zs
    #         dd_out["h_pv"] = h_pv
    #
    #         peak.numpy_ttheta = tth_hkl + setup.offset_ttheta
    #         peak.width_ttheta = h_pv
    #         peak.numpy_to_items()
    #
    #         # texture
    #         if texture is not None:
    #             cond_1 = dd_in is not None
    #             cond_2 = ((not(setup.offset_phi_refinement)))
    #             if cond_1 & cond_2:
    #                 cos_alpha_ang_3d = dd_in["cos_alpha_ang_3d"]
    #                 sin_alpha_ang_3d = dd_in["sin_alpha_ang_3d"]
    #             else:
    #                 cos_alpha_ang_3d = cos_theta_3d * sin_phi_3d
    #                 sin_alpha_ang_3d = numpy.sqrt(1.-cos_alpha_ang_3d**2)
    #             dd_out["cos_alpha_ang_3d"] = cos_alpha_ang_3d
    #             dd_out["sin_alpha_ang_3d"] = sin_alpha_ang_3d
    #             cond_2 = (cond_2 & (not(texture.is_variables())) &
    #                       (not(cell.is_variables())))
    #             if cond_1 & cond_2:
    #                 texture_3d = dd_in["texture_3d"]
    #             else:
    #                 cos_alpha_ax = calc_cos_ang(cell, h_ax, k_ax, l_ax,
    #                                             index_h, index_k, index_l)
    #
    #                 c_help = 1.-cos_alpha_ax**2
    #                 c_help[c_help < 0.] = 0.
    #                 sin_alpha_ax = numpy.sqrt(c_help)
    #                 cos_alpha_ax_3d = numpy.meshgrid(tth, phi, cos_alpha_ax,
    #                                                  indexing="ij")[2]
    #                 sin_alpha_ax_3d = numpy.meshgrid(tth, phi, sin_alpha_ax,
    #                                                  indexing="ij")[2]
    #                 cos_alpha_3d = cos_alpha_ax_3d*cos_alpha_ang_3d + \
    #                     sin_alpha_ax_3d*sin_alpha_ang_3d
    #                 texture_3d = g_2 + (1.-g_2) * \
    #                     (1./g_1 + (g_1**2-1./g_1)*cos_alpha_3d**2)**(-1.5)
    #             dd_out["texture_3d"] = texture_3d
    #
    #             profile_3d = profile_3d*texture_3d
    #
    #         res_u_3d = profile_3d*iint_u_3d
    #         res_d_3d = profile_3d*iint_d_3d
    #
    #         # 0.5 to have the same meaning for scale factor as in FullProf
    #         res_u_2d += 0.5*phase_scale*res_u_3d.sum(axis=2)
    #         res_d_2d += 0.5*phase_scale*res_d_3d.sum(axis=2)
    #         l_peak.append(peak)
    #         l_dd_out.append(dd_out)
    #
    #     proc.ttheta_corrected = tth_zs
    #     proc.intensity_plus_net = res_u_2d
    #     proc.intensity_minus_net = res_d_2d
    #     proc.intensity_plus_total = res_u_2d+int_bkgd
    #     proc.intensity_minus_total = res_d_2d+int_bkgd
    #     if flag_internal:
    #         if background.is_variables():
    #             background.form_ttheta_phi_intensity()
    #         proc.form_ttheta_phi_intensity_plus_net()
    #         proc.form_ttheta_phi_intensity_minus_net()
    #         proc.form_ttheta_phi_intensity_plus_total()
    #         proc.form_ttheta_phi_intensity_minus_total()
    #     l_calc_objs = l_refln + l_refln_s + l_peak
    #     l_calc_objs.append(proc)
    #     self.add_items(l_calc_objs)
    #
    #     return proc, l_peak, l_refln, l_dd_out
    #
    # def calc_chi_sq(self, l_crystal, flag_internal: bool = True):
    #     """
    #     Calculate chi square.
    #
    #     Arguments
    #     ---------
    #         - l_crystal: a list of Crystal objects of cryspy library
    #         - flag_internal: a flag to calculate internal objects
    #           (default is True)
    #
    #     Output arguments
    #     ----------------
    #         - chi_sq_val: chi square of flip ratio
    #           (Sum_i ((y_e_i - y_m_i) / sigma_i)**2)
    #         - n: number of measured reflections
    #     """
    #     meas = self.pd2d_meas
    #
    #     tth = meas.ttheta
    #     phi = meas.phi
    #     int_u_exp = meas.intensity_plus
    #     sint_u_exp = meas.intensity_plus_sigma
    #     int_d_exp = meas.intensity_minus
    #     sint_d_exp = meas.intensity_minus_sigma
    #
    #     l_peak_in, l_refln_in = [], []
    #     l_refln_susceptibility_in = []
    #     l_dd_in = []
    #     flag_1 = not(flag_internal)
    #
    #     try:
    #         dd = self.dd
    #         flag_2 = True
    #     except AttributeError:
    #         flag_2 = False
    #     if (flag_1 & flag_2):
    #         for phase_item in self.phase.items:
    #             crystal = None
    #             for cryst in l_crystal:
    #                 if cryst.data_name.lower() == phase_item.label.lower():
    #                     crystal = cryst
    #                     break
    #             attr_peak = f"pd2d_peak_{crystal.data_name:}"
    #             attr_refln = f"refln_{crystal.data_name:}"
    #             attr_refln_s = f"refln_susceptibility_{crystal.data_name:}"
    #             l_peak_in.append(getattr(self, attr_peak))
    #             l_refln_in.append(getattr(self, attr_refln))
    #             l_refln_susceptibility_in.append(getattr(self, attr_refln_s))
    #
    #     cond_tth_in = numpy.ones(tth.size, dtype=bool)
    #     cond_phi_in = numpy.ones(phi.size, dtype=bool)
    #     try:
    #         range_ = self.range
    #         cond_tth_in = numpy.logical_and(cond_tth_in, tth >=
    #                                         range_.ttheta_min)
    #         cond_tth_in = numpy.logical_and(cond_tth_in, tth <=
    #                                         range_.ttheta_max)
    #
    #         cond_phi_in = numpy.logical_and(cond_phi_in, phi >= range_.phi_min)
    #         cond_phi_in = numpy.logical_and(cond_phi_in, phi <= range_.phi_max)
    #
    #     except AttributeError:
    #         pass
    #     # cond_1_in, cond_2_in = numpy.meshgrid(cond_tth_in, cond_phi_in,
    #     #                                       indexing="ij")
    #     # cond_in = numpy.logical_and(cond_1_in, cond_2_in)
    #     tth_in = tth[cond_tth_in]
    #     phi_in = phi[cond_phi_in]
    #     int_u_exp_in = int_u_exp[cond_tth_in, :][:, cond_phi_in]
    #     sint_u_exp_in = sint_u_exp[cond_tth_in, :][:, cond_phi_in]
    #     int_d_exp_in = int_d_exp[cond_tth_in, :][:, cond_phi_in]
    #     sint_d_exp_in = sint_d_exp[cond_tth_in, :][:, cond_phi_in]
    #
    #     proc, l_peak, l_refln, l_dd_out = self.calc_profile(
    #         tth_in, phi_in, l_crystal, l_peak_in=l_peak_in,
    #         l_refln_in=l_refln_in,
    #         l_refln_susceptibility_in=l_refln_susceptibility_in,
    #         l_dd_in=l_dd_in, flag_internal=flag_internal)
    #     proc.intensity_plus = int_u_exp_in
    #     proc.intensity_plus_sigma = sint_u_exp_in
    #     proc.intensity_minus = int_d_exp_in
    #     proc.intensity_minus_sigma = sint_d_exp_in
    #
    #     # self.proc = proc
    #     # self.peaks = l_peak
    #     # self.reflns = l_refln
    #     self.dd = l_dd_out
    #
    #     int_u_mod = proc.intensity_plus_total
    #     int_d_mod = proc.intensity_minus_total
    #
    #     sint_sum_exp_in = (sint_u_exp_in**2 + sint_d_exp_in**2)**0.5
    #
    #     chi_sq_u = ((int_u_mod-int_u_exp_in)/sint_u_exp_in)**2
    #     chi_sq_d = ((int_d_mod-int_d_exp_in)/sint_d_exp_in)**2
    #     chi_sq_sum = ((int_u_mod+int_d_mod-int_u_exp_in-int_d_exp_in) /
    #                   sint_sum_exp_in)**2
    #     chi_sq_dif = ((int_u_mod-int_d_mod-int_u_exp_in+int_d_exp_in) /
    #                   sint_sum_exp_in)**2
    #
    #     cond_u = numpy.logical_not(numpy.isnan(chi_sq_u))
    #     cond_d = numpy.logical_not(numpy.isnan(chi_sq_d))
    #     cond_sum = numpy.logical_not(numpy.isnan(chi_sq_sum))
    #     cond_dif = numpy.logical_not(numpy.isnan(chi_sq_dif))
    #
    #     # exclude region
    #     try:
    #         exclude = self.exclude
    #         l_excl_tth_min = exclude.numpy_ttheta_min
    #         l_excl_tth_max = exclude.numpy_ttheta_max
    #         l_excl_phi_min = exclude.numpy_phi_min
    #         l_excl_phi_max = exclude.numpy_phi_max
    #         for excl_tth_min, excl_tth_max, excl_phi_min, excl_phi_max in \
    #             zip(l_excl_tth_min, l_excl_tth_max, l_excl_phi_min,
    #                 l_excl_phi_max):
    #             cond_1 = numpy.logical_or(tth_in < 1.*excl_tth_min,
    #                                       tth_in > 1.*excl_tth_max)
    #             cond_2 = numpy.logical_or(phi_in < 1.*excl_phi_min,
    #                                       phi_in > 1.*excl_phi_max)
    #             cond_11, cond_22 = numpy.meshgrid(cond_1, cond_2,
    #                                               indexing="ij")
    #             cond_12 = numpy.logical_or(cond_11, cond_22)
    #             cond_u = numpy.logical_and(cond_u, cond_12)
    #             cond_d = numpy.logical_and(cond_d, cond_12)
    #             cond_sum = numpy.logical_and(cond_sum, cond_12)
    #     except AttributeError:
    #         pass
    #
    #     chi_sq_u_val = (chi_sq_u[cond_u]).sum()
    #     n_u = cond_u.sum()
    #
    #     chi_sq_d_val = (chi_sq_d[cond_d]).sum()
    #     n_d = cond_d.sum()
    #
    #     chi_sq_sum_val = (chi_sq_sum[cond_sum]).sum()
    #     n_sum = cond_sum.sum()
    #
    #     chi_sq_dif_val = (chi_sq_dif[cond_dif]).sum()
    #     n_dif = cond_dif.sum()
    #     chi2 = self.chi2
    #     flag_u = chi2.up
    #     flag_d = chi2.down
    #     flag_sum = chi2.sum
    #     flag_dif = chi2.diff
    #
    #     chi_sq_val = (int(flag_u)*chi_sq_u_val + int(flag_d)*chi_sq_d_val +
    #                   int(flag_sum)*chi_sq_sum_val +
    #                   int(flag_dif)*chi_sq_dif_val)
    #     n = (int(flag_u)*n_u + int(flag_d)*n_d + int(flag_sum)*n_sum +
    #          int(flag_dif)*n_dif)
    #     # print(f"chi_sq_val/n: {chi_sq_val/n:.2f}    \
    #     #     chi_sq_val: {chi_sq_val: .2f}")
    #     # d_exp_out = {"chi_sq_val": chi_sq_val, "n": n}
    #     # d_exp_out.update(d_exp_prof_out)
    #
    #     if flag_internal:
    #         refine_ls = RefineLs(number_reflns=n,
    #                              goodness_of_fit_all=chi_sq_val/float(n),
    #                              weighting_scheme="sigma")
    #         self.refine_ls = refine_ls
    #         proc.form_ttheta_phi_intensity_bkg_calc()
    #         proc.form_ttheta_phi_intensity_plus()
    #         proc.form_ttheta_phi_intensity_plus_sigma()
    #         proc.form_ttheta_phi_intensity_minus()
    #         proc.form_ttheta_phi_intensity_minus_sigma()
    #     return chi_sq_val, n
    #
    # def calc_for_iint(self, index_h, index_k, index_l, crystal,
    #                   flag_internal: bool = True):
    #     """Calculate the integral intensity for h, k, l reflections."""
    #     index_hkl = numpy.stack([index_h, index_k, index_l], axis=0)
    #     setup = self.setup
    #     field = float(setup.field)
    #
    #     refln = crystal.calc_refln(index_hkl)
    #     f_nucl = refln.numpy_f_calc
    #
    #     f_nucl_sq = abs(f_nucl*f_nucl.conjugate())
    #
    #     refln_s = crystal.calc_refln_susceptibility(index_hkl)
    #     sft_11 = refln_s.numpy_chi_11_calc
    #     sft_12 = refln_s.numpy_chi_12_calc
    #     sft_13 = refln_s.numpy_chi_13_calc
    #     sft_21 = refln_s.numpy_chi_21_calc
    #     sft_22 = refln_s.numpy_chi_22_calc
    #     sft_23 = refln_s.numpy_chi_23_calc
    #     sft_31 = refln_s.numpy_chi_31_calc
    #     sft_32 = refln_s.numpy_chi_32_calc
    #     sft_33 = refln_s.numpy_chi_33_calc
    #     _ij = numpy.stack([sft_11, sft_12, sft_13, sft_21, sft_22, sft_23, sft_31, sft_32, sft_33], axis=0)
    #
    #     cell = crystal.cell
    #     unit_cell_parameters = cell.get_unit_cell_parameters()
    #     t_ij = calc_matrix_t(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
    #     #SIGMA = T CHI T^-1 = T chi T^T (because T is rotation matrix, therefore T^-1 = T^T)
    #     th_ij = calc_m1_m2_m1t(t_ij, _ij)[0]
    #     th_11, th_12, th_13, th_22, th_23 = th_ij[0], th_ij[1], th_ij[2], th_ij[4], th_ij[5]
    #
    #     # f_m_p_sin_sq = (field**2)*abs(0.5*(th_11*th_11.conjugate()+\
    #     #                th_22*th_22.conjugate())+th_12*th_12.conjugate())
    #     # f_m_p_cos_sq = (field**2)*abs(th_13*th_13.conjugate()+\
    #     #                 th_23*th_23.conjugate())
    #     # f_m_p_field = 0.5*field*(th_11+th_22)
    #
    #     f_m_p_sin_sq = abs(0.5 * (th_11 * th_11.conjugate()+th_22 *
    #                               th_22.conjugate())+th_12 * th_12.conjugate())
    #     f_m_p_cos_sq = abs(th_13 * th_13.conjugate() +
    #                        th_23 * th_23.conjugate())
    #     f_m_p_field = 0.5 * (th_11+th_22)
    #     cross_sin = 2. * (f_nucl.real * f_m_p_field.real +
    #                       f_nucl.imag * f_m_p_field.imag)
    #
    #     return f_nucl_sq, f_m_p_sin_sq, f_m_p_cos_sq, cross_sin, refln, refln_s
    #
    # def _gauss_pd(self, tth_2d):
    #     """One dimensional gauss powder diffraction."""
    #     ag, bg = self.ag, self.bg
    #     val_1 = bg*tth_2d**2
    #     val_2 = numpy.where(val_1 < 5., numpy.exp(-val_1), 0.)
    #     self.gauss_pd = ag*val_2
    #
    # def _lor_pd(self, tth_2d):
    #     """One dimensional lorentz powder diffraction."""
    #     al, bl = self.al, self.bl
    #     self.lor_pd = al*1./(1.+bl*tth_2d**2)
    #
    # def calc_shape_profile(
    #         self, tth, phi, tth_hkl, phase_igsize: float = 0.,
    #         phase_u: float = 0., phase_v: float = 0., phase_w: float = 0.,
    #         phase_x: float = 0., phase_y: float = 0.):
    #     """
    #     Calculate profile in the range ttheta.
    #
    #     For reflections placed on
    #     ttheta_hkl with i_g parameter by default equal to zero
    #
    #     tth, phi, tth_hkl in degrees
    #     """
    #     setup = self.setup
    #     zero_shift = float(setup.offset_ttheta)
    #     tth_zs = tth-zero_shift
    #
    #     resolution = self.pd2d_instr_resolution
    #
    #     h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l = resolution.calc_resolution(
    #         tth_hkl, phase_igsize=phase_igsize, phase_u=phase_u,
    #         phase_v=phase_v, phase_w=phase_w, phase_x=phase_x, phase_y=phase_y)
    #
    #     tth_3d, phi_3d, tth_hkl_3d = numpy.meshgrid(tth_zs, phi, tth_hkl,
    #                                                 indexing="ij")
    #
    #     self.ag = numpy.meshgrid(tth_zs, phi, a_g, indexing="ij")[2]
    #     self.bg = numpy.meshgrid(tth_zs, phi, b_g, indexing="ij")[2]
    #     self.al = numpy.meshgrid(tth_zs, phi, a_l, indexing="ij")[2]
    #     self.bl = numpy.meshgrid(tth_zs, phi, b_l, indexing="ij")[2]
    #     eta_3d = numpy.meshgrid(tth_zs, phi, eta, indexing="ij")[2]
    #     self.eta = eta_3d
    #
    #     self._gauss_pd(tth_3d-tth_hkl_3d)
    #     self._lor_pd(tth_3d-tth_hkl_3d)
    #     g_pd2d_3d = self.gauss_pd
    #     l_pd2d_3d = self.lor_pd
    #
    #     np_shape_3d = eta_3d * l_pd2d_3d + (1.-eta_3d) * g_pd2d_3d
    #
    #     asymmetry = self.pd2d_instr_reflex_asymmetry
    #     np_ass_2d = asymmetry.calc_asymmetry(tth_zs, tth_hkl, h_pv)
    #     np_ass_3d = np_ass_2d[:, numpy.newaxis, :] * numpy.ones(
    #         phi.size, dtype=float)[numpy.newaxis, :, numpy.newaxis]
    #
    #     # Lorentz factor
    #     tth_rad = tth_zs*numpy.pi/180.
    #     np_lor_1d = 1./(numpy.sin(tth_rad)*numpy.sin(0.5*tth_rad))
    #     np_lor_3d = numpy.meshgrid(np_lor_1d, phi, tth_hkl, indexing="ij")[0]
    #
    #     profile_3d = np_shape_3d*np_ass_3d*np_lor_3d
    #
    #     return profile_3d, tth_zs, h_pv
    #
    # def params_to_cif(self, separator="_", flag: bool = False,
    #                   flag_minimal: bool = True) -> str:
    #     """Save parameters to cif format."""
    #     ls_out = []
    #     l_cls = (Pd2dBackground, Pd2dInstrResolution, PhaseL, DiffrnRadiation,
    #              Setup, Range, Chi2, Extinction, ExcludeL,
    #              Pd2dInstrReflexAsymmetry, TextureL)
    #     l_obj = [item for item in self.items if type(item) in l_cls]
    #     l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
    #     l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
    #     ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
    #     ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
    #     return "\n".join(ls_out)
    #
    # def data_to_cif(self, separator="_", flag: bool = False,
    #                 flag_minimal: bool = True) -> str:
    #     """Save data to cif format."""
    #     ls_out = []
    #     l_cls = (Pd2dMeas, )
    #     l_obj = [item for item in self.items if type(item) in l_cls]
    #     l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
    #     l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
    #     ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
    #     ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
    #     return "\n".join(ls_out)
    #
    # def calc_to_cif(self, separator="_", flag=False, flag_minimal=True) -> str:
    #     """Save calculations to cif format."""
    #     ls_out = []
    #     l_cls = (RefineLs, ReflnL, ReflnSusceptibilityL, Pd2dPeakL, Pd2dProc)
    #     l_obj = [item for item in self.items if type(item) in l_cls]
    #     l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
    #     l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
    #     ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
    #     ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
    #     return "\n".join(ls_out)

    def apply_constraints(self):
        """Apply constraints."""
        pass

    def plots(self):
        if self.is_attribute("pd2d_proc"):
            pd2d_proc = self.pd2d_proc
            fig = pd2d_proc.plot_gamma_nu()[0]
            l_pd_peak = [item for item in self.items if isinstance(item, PdPeakL)]
            [ax1, ax2] = fig.axes[:2]
            gamma_min, gamma_max = ax1.get_xlim()
            nu_min, nu_max = ax1.get_ylim()
            for pd_peak in l_pd_peak:
                if pd_peak.is_attribute("gamma"):
                    index_h = numpy.array(pd_peak.index_h, dtype=int)
                    index_k = numpy.array(pd_peak.index_k, dtype=int)
                    index_l = numpy.array(pd_peak.index_l, dtype=int)
                    gamma = numpy.array(pd_peak.gamma, dtype=float)
                    nu = numpy.array(pd_peak.nu, dtype=float)

                    flag_gn = numpy.logical_and(
                        numpy.logical_and(gamma>=gamma_min, gamma<=gamma_max),
                        numpy.logical_and(nu>=nu_min, nu<=nu_max))

                    if (pd_peak.is_attribute("intensity_plus") and pd_peak.is_attribute("intensity_minus")):
                        i_plus = numpy.array(pd_peak.intensity_plus, dtype=float)
                        i_minus = numpy.array(pd_peak.intensity_minus, dtype=float)
                        inv_lf = numpy.sin(gamma*numpy.pi/180)*numpy.sin(0.5*gamma*numpy.pi/180)
                        i_sum = numpy.abs(i_plus*flag_gn+i_minus*flag_gn)
                        i_difference = numpy.abs(i_plus*flag_gn-i_minus*flag_gn)
                        flag_i_plus = (i_sum-i_sum.min()) > 0.01 * (i_sum.max() - i_sum.min())
                        flag_i_difference = i_difference  > 0.01 * i_difference.max()
                        flag_i = numpy.logical_or(flag_i_plus, flag_i_difference)
                        flag_gn = numpy.logical_and(flag_gn, flag_i)

                    ax1.plot(gamma[flag_gn], 0.5*(nu[flag_gn]+nu_max), "k.", alpha=0.3)
                    ax2.plot(gamma[flag_gn], 0.5*(nu[flag_gn]+nu_max), "k.", alpha=0.3)
                    ax1.plot(gamma[flag_gn], nu_min+0.5*(-nu[flag_gn]+nu_max), "k.", alpha=0.3)
                    ax2.plot(gamma[flag_gn], nu_min+0.5*(-nu[flag_gn]+nu_max), "k.", alpha=0.3)

                    if (pd_peak.is_attribute("intensity_plus") and pd_peak.is_attribute("intensity_minus")):
                        i_plus = numpy.array(pd_peak.intensity_plus, dtype=float)
                        i_minus = numpy.array(pd_peak.intensity_minus, dtype=float)
                        inv_lf = numpy.sin(gamma*numpy.pi/180)*numpy.sin(0.5*gamma*numpy.pi/180)
                        i_sum = numpy.abs(i_plus*flag_gn+i_minus*flag_gn)
                        i_difference = numpy.abs(i_plus*flag_gn-i_minus*flag_gn)
                        flag_i_plus = (i_sum-i_sum.min()) > 0.2 * (i_sum.max() - i_sum.min())
                        flag_i_difference = i_difference  > 0.2 * i_difference.max()
                        flag_i = numpy.logical_or(flag_i_plus, flag_i_difference)
                        flag_gn = numpy.logical_and(flag_gn, flag_i)

                    l_gn = []
                    for i_h, i_k, i_l, i_g, i_n in zip(
                            index_h[flag_gn], index_k[flag_gn], index_l[flag_gn], gamma[flag_gn], nu[flag_gn]):
                        if not((i_g, i_n) in l_gn):
                            s_text = f"({i_h:}{i_k:}{i_l:})"
                            ax1.text(i_g, 0.5*(+i_n+nu_max), s_text, alpha=0.5, color="k")
                            ax2.text(i_g, 0.5*(+i_n+nu_max), s_text, alpha=0.5, color="k")
                            ax1.text(i_g, 0.5*(-i_n+nu_max)+nu_min, s_text, alpha=0.5, color="k")
                            ax2.text(i_g, 0.5*(-i_n+nu_max)+nu_min, s_text, alpha=0.5, color="k")
                            l_gn.append((i_g, i_n))
            return [(fig, ax1), ]
            if self.is_attribute("chi2"):
                flag_up = self.chi2.up
                flag_down = self.chi2.down
                flag_sum = self.chi2.sum
                flag_diff = self.chi2.diff
            else:
                flag_up, flag_down, flag_sum = False, False, True
                flag_diff = False

            if flag_sum:
                fig_s, ax_s = pd2d_proc.plot_projection_sum()
                ax_s.set_title(self.data_name + " - "+ax_s.title.get_text())
                y_min_s, y_max_s = ax_s.get_ylim()
                y_dist_s = y_max_s-y_min_s
                y_step_s = 0.

            if flag_diff:
                fig_d, ax_d = pd2d_proc.plot_projection_diff()
                ax_d.set_title(self.data_name + " - "+ax_d.title.get_text())
                y_min_d, y_max_d = ax_d.get_ylim()
                y_dist_d = y_max_d-y_min_d
                y_step_d = 0.

            for item in self.items:
                if isinstance(item, Pd2dPeakL):
                    np_tth = item.numpy_ttheta
                    if flag_sum:
                        ax_s.plot(np_tth, 0.*np_tth+y_min_s-y_step_s, "|",
                                  label=item.loop_name)
                        y_step_s += 0.05*y_dist_s
                    if flag_diff:
                        ax_d.plot(np_tth, 0.*np_tth+y_min_d-y_step_d, "|",
                                  label=item.loop_name)
                        y_step_d += 0.05*y_dist_d
            res = []
            if flag_sum:
                ax_s.legend(loc='upper right')
                res.append((fig_s, ax_s))
            if flag_diff:
                ax_d.legend(loc='upper right')
                res.append((fig_d, ax_d))
            return res
        elif self.is_attribute("pd2d_meas"):
            return self.pd2d_meas.plots()
        return []

    def get_dictionary(self):
        """Form dictionary. See documentation moduel CrysPy using Jupyter notebook.
        """
        ddict = super(Pd2d, self).get_dictionary()
        pd2d_meas, range_, exclude = None, None, None
        # self.form_object()
        # ddict = {}
        # setup, pd2d_meas, resolution = None, None, None
        # pd2d_background, range_, exclude = None, None, None
        # asymmetry, diffrn_radiation = None, None
        # phase, texture, chi2 = None, None, None
        #
        # l_obj = take_items_by_class(self, (Setup, ))
        # if len(l_obj) > 0:
        #     setup = l_obj[0]

        l_obj = take_items_by_class(self, (Pd2dMeas, ))
        if len(l_obj) > 0:
            pd2d_meas = l_obj[0]

        # l_obj = take_items_by_class(self, (Pd2dInstrResolution, ))
        # if len(l_obj) > 0:
        #     resolution = l_obj[0]
        #
        # l_obj = take_items_by_class(self, (Pd2dBackground, ))
        # if len(l_obj) > 0:
        #     pd2d_background = l_obj[0]

        l_obj = take_items_by_class(self, (Range, ))
        if len(l_obj) > 0:
            range_ = l_obj[0]

        l_obj = take_items_by_class(self, (ExcludeL, ))
        if len(l_obj) > 0:
            exclude = l_obj[0]

        # l_obj = take_items_by_class(self, (Pd2dInstrReflexAsymmetry, ))
        # if len(l_obj) > 0:
        #     asymmetry = l_obj[0]
        #
        # l_obj = take_items_by_class(self, (DiffrnRadiation, ))
        # if len(l_obj) > 0:
        #     diffrn_radiation = l_obj[0]
        #
        # l_obj = take_items_by_class(self, (PhaseL, ))
        # if len(l_obj) > 0:
        #     phase = l_obj[0]
        #
        # l_obj = take_items_by_class(self, (TextureL, ))
        # if len(l_obj) > 0:
        #     texture = l_obj[0]
        #
        # l_obj = take_items_by_class(self, (Chi2, ))
        # if len(l_obj) > 0:
        #     chi2 = l_obj[0]
        #
        # ddict["name"] = self.data_name
        # ddict["type_name"] = self.get_name()
        # if setup is not None:
        #     ddict["magnetic_field"] = numpy.array([setup.field], dtype=float)
        #     ddict["wavelength"] = numpy.array([setup.wavelength], dtype=float)
        #     ddict["flags_wavelength"] = numpy.array([setup.wavelength_refinement], dtype=bool)
        #     ddict["offset_gamma"] = numpy.array([setup.offset_gamma * numpy.pi/180.], dtype=float) # FIXME
        #     ddict["flags_offset_gamma"] = numpy.array([setup.offset_gamma_refinement], dtype=bool) # FIXME
        #     ddict["offset_nu"] = numpy.array([setup.offset_nu * numpy.pi/180.], dtype=float) # FIXME
        #     ddict["flags_offset_nu"] = numpy.array([setup.offset_nu_refinement], dtype=bool) # FIXME

        # if chi2 is not None:
        #     ddict["flag_chi_sq_sum"] = chi2.sum
        #     ddict["flag_chi_sq_difference"] = chi2.diff

        if pd2d_meas is not None:
            gamma_deg = numpy.array(pd2d_meas.gamma, dtype=float) #
            gamma_min_deg = range_.gamma_min
            gamma_max_deg = range_.gamma_max
            nu_deg = numpy.array(pd2d_meas.nu, dtype=float) #
            nu_min_deg = range_.nu_min
            nu_max_deg = range_.nu_max

            flag_gamma_in = numpy.logical_and(
                gamma_deg >= gamma_min_deg,
                gamma_deg <= gamma_max_deg)

            flag_nu_in = numpy.logical_and(
                nu_deg >= nu_min_deg,
                nu_deg <= nu_max_deg)

            flag_in = flag_gamma_in[:, na] * flag_nu_in[na, :]

            # ddict["ttheta"] = ttheta_deg[flag_in] * numpy.pi/180.
            ddict["gamma"] = gamma_deg[flag_gamma_in] * numpy.pi/180.
            ddict["nu"] = nu_deg[flag_nu_in] * numpy.pi/180.

            if pd2d_meas.is_attribute("gn_intensity_plus"):
                int_plus = numpy.array(pd2d_meas.gn_intensity_plus, dtype=float)[flag_gamma_in][:, flag_nu_in]
                s_int_plus = numpy.array(pd2d_meas.gn_intensity_plus_sigma, dtype=float)[flag_gamma_in][:, flag_nu_in]
                int_minus = numpy.array(pd2d_meas.gn_intensity_minus, dtype=float)[flag_gamma_in][:, flag_nu_in]
                s_int_minus = numpy.array(pd2d_meas.gn_intensity_minus_sigma, dtype=float)[flag_gamma_in][:, flag_nu_in]
                ddict["signal_exp_plus"] = numpy.stack([int_plus, s_int_plus], axis=0)
                ddict["signal_exp_minus"] = numpy.stack([int_minus, s_int_minus], axis=0)
            else:
                int_sum = numpy.array(pd2d_meas.gn_intensity, dtype=float)[flag_gamma_in][:, flag_nu_in]
                s_int_sum = numpy.array(pd2d_meas.gn_intensity_sigma, dtype=float)[flag_gamma_in][:, flag_nu_in]
                ddict["signal_exp"] = numpy.stack([int_sum, s_int_sum], axis=0)

            gamma_in_range = gamma_deg[flag_gamma_in]
            nu_in_range = nu_deg[flag_nu_in]
            ttheta_in_range, phi_in_range = calc_ttheta_phi_by_gamma_nu(
                gamma_in_range[:, na]*numpy.pi/180. - ddict["offset_gamma"],
                nu_in_range[na, :]*numpy.pi/180. - ddict["offset_nu"],
                flag_gamma=False, flag_nu=False)[:2]
            ttheta_in_range = ttheta_in_range*180./numpy.pi
            phi_in_range = phi_in_range*180./numpy.pi
            flag_exclude = numpy.zeros(gamma_in_range.shape + nu_in_range.shape, dtype=bool)
            if exclude is not None:
                for item_e in exclude.items:
                    if item_e.is_attribute("gamma_min"):
                        if item_e.gamma_min is not None:
                            flag_gamma_in_1 = numpy.logical_and(
                                gamma_in_range >= item_e.gamma_min,
                                gamma_in_range <= item_e.gamma_max)
                            flag_nu_in_1 = numpy.logical_and(
                                nu_in_range >= item_e.nu_min,
                                nu_in_range <= item_e.nu_max)
                            flag_in_1 = flag_gamma_in_1[:, na] * flag_nu_in_1[na, :]
                            flag_exclude = numpy.logical_or(flag_exclude, flag_in_1)
                    if item_e.is_attribute("ttheta_min"):
                        if item_e.ttheta_min is not None:
                            flag_ttheta_in_1 = numpy.logical_and(
                                ttheta_in_range >= item_e.ttheta_min,
                                ttheta_in_range <= item_e.ttheta_max)
                            flag_phi_in_1 = numpy.logical_and(
                                phi_in_range >= item_e.phi_min,
                                phi_in_range <= item_e.phi_max)
                            flag_in_1 = numpy.logical_and(
                                flag_ttheta_in_1,
                                flag_phi_in_1)
                            flag_exclude = numpy.logical_or(flag_exclude, flag_in_1)

            ddict["excluded_points"] = flag_exclude

        # if resolution is not None:
        #     ddict["resolution_parameters"] = numpy.array([
        #         resolution.u, resolution.v, resolution.w,
        #         resolution.x, resolution.y], dtype=float)
        #
        #     ddict["flags_resolution_parameters"] = numpy.array([
        #         resolution.u_refinement, resolution.v_refinement, resolution.w_refinement,
        #         resolution.x_refinement, resolution.y_refinement], dtype=bool)
        #
        #     ddict["resolution_phi_parameter"] = numpy.array([
        #         resolution.phi,], dtype=float)
        #     ddict["flags_resolution_phi_parameter"] = numpy.array([
        #         resolution.phi_refinement,], dtype=float)
        #
        # if asymmetry is not None:
        #     ddict["asymmetry_parameters"] = numpy.array([
        #         asymmetry.p1, asymmetry.p2, asymmetry.p3,
        #         asymmetry.p4], dtype=float)
        #
        #     ddict["flags_asymmetry_parameters"] = numpy.array([
        #         asymmetry.p1_refinement, asymmetry.p2_refinement, asymmetry.p3_refinement,
        #         asymmetry.p4_refinement], dtype=bool)
        # else:
        #     ddict["asymmetry_parameters"] = numpy.array([
        #         0, 0, 0, 0], dtype=float)
        #
        #     ddict["flags_asymmetry_parameters"] = numpy.array([
        #         False, False, False, False], dtype=bool)
        #
        # if diffrn_radiation is not None:
        #     beam_polarization = diffrn_radiation.polarization
        #     flipper_efficiency = diffrn_radiation.efficiency
        #     ddict["beam_polarization"] = numpy.array([beam_polarization], dtype=float)
        #     ddict["flipper_efficiency"] = numpy.array([flipper_efficiency], dtype=float)
        #     ddict["flags_beam_polarization"] = numpy.array([diffrn_radiation.polarization_refinement], dtype=bool)
        #     ddict["flags_flipper_efficiency"] = numpy.array([diffrn_radiation.efficiency_refinement], dtype=bool)
        #
        # if phase is not None:
        #     ddict["phase_name"] = numpy.array(phase.label, dtype=str)
        #     if phase.is_attribute("u"):
        #         p_u = numpy.array(phase.u, dtype=float)
        #         r_u = numpy.array(phase.u_refinement, dtype=bool)
        #     else:
        #         p_u = numpy.zeros((len(phase.items),), dtype=float)
        #         r_u = numpy.zeros((len(phase.items),), dtype=bool)
        #
        #     if phase.is_attribute("v"):
        #         p_v = numpy.array(phase.v, dtype=float)
        #         r_v = numpy.array(phase.v_refinement, dtype=bool)
        #     else:
        #         p_v = numpy.zeros((len(phase.items),), dtype=float)
        #         r_v = numpy.zeros((len(phase.items),), dtype=bool)
        #
        #     if phase.is_attribute("w"):
        #         p_w = numpy.array(phase.w, dtype=float)
        #         r_w = numpy.array(phase.w_refinement, dtype=bool)
        #     else:
        #         p_w = numpy.zeros((len(phase.items),), dtype=float)
        #         r_w = numpy.zeros((len(phase.items),), dtype=bool)
        #
        #     if phase.is_attribute("x"):
        #         p_x = numpy.array(phase.x, dtype=float)
        #         r_x = numpy.array(phase.x_refinement, dtype=bool)
        #     else:
        #         p_x = numpy.zeros((len(phase.items),), dtype=float)
        #         r_x = numpy.zeros((len(phase.items),), dtype=bool)
        #
        #     if phase.is_attribute("y"):
        #         p_y = numpy.array(phase.y, dtype=float)
        #         r_y = numpy.array(phase.y_refinement, dtype=bool)
        #     else:
        #         p_y = numpy.zeros((len(phase.items),), dtype=float)
        #         r_y = numpy.zeros((len(phase.items),), dtype=bool)
        #
        #     ddict["phase_resolution_parameters"] = numpy.stack([p_u, p_v, p_w, p_x, p_y], axis=0)
        #     ddict["flags_phase_resolution_parameters"] = numpy.stack([r_u, r_v, r_w, r_x, r_y], axis=0)
        #
        #     if phase.is_attribute("igsize"):
        #         ddict["phase_ig"] = numpy.array(phase.igsize, dtype=float)
        #         ddict["flags_phase_ig"] = numpy.array(phase.igsize_refinement, dtype=bool)
        #     else:
        #         ddict["phase_ig"] = numpy.zeros((len(phase.items),), dtype=float)
        #         ddict["flags_phase_ig"] = numpy.zeros((len(phase.items),), dtype=bool)
        #
        #     if phase.is_attribute("scale"):
        #         ddict["phase_scale"] = numpy.array(phase.scale, dtype=float)
        #         ddict["flags_phase_scale"] = numpy.array(phase.scale_refinement, dtype=bool)
        #     else:
        #         ddict["phase_scale"] = numpy.zeros((len(phase.items),), dtype=float)
        #         ddict["flags_phase_scale"] = numpy.zeros((len(phase.items),), dtype=bool)
        #
        # if texture is not None:
        #     ddict["texture_name"] = numpy.array(texture.label, dtype=str)
        #     ddict["texture_g1"] = numpy.array(texture.g_1, dtype=float)
        #     ddict["texture_g2"] = numpy.array(texture.g_2, dtype=float)
        #     ddict["texture_axis"] = numpy.array(
        #         [texture.h_ax, texture.k_ax, texture.l_ax], dtype=float)
        #     ddict["flags_texture_g1"] = numpy.array(texture.g_1_refinement, dtype=bool)
        #     ddict["flags_texture_g2"] = numpy.array(texture.g_2_refinement, dtype=bool)
        #     ddict["flags_texture_axis"] = numpy.array(
        #         [texture.h_ax_refinement,
        #          texture.k_ax_refinement,
        #          texture.l_ax_refinement], dtype=bool)
        #
        # if pd2d_background is not None:
        #     ddict["background_gamma"] = numpy.array(pd2d_background.gamma, dtype=float)*numpy.pi/180.
        #     ddict["background_nu"] = numpy.array(pd2d_background.nu, dtype=float)*numpy.pi/180.
        #     ddict["background_intensity"] = numpy.array(pd2d_background.intensity, dtype=float)
        #     ddict["flags_background_intensity"] = numpy.array(pd2d_background.intensity_refinement, dtype=bool)

        return ddict

    def take_parameters_from_dictionary(self, ddict_diffrn, l_parameter_name: list=None, l_sigma: list=None):
        keys = ddict_diffrn.keys()
        if l_parameter_name is not None:
            parameter_label = [hh[0] for hh in l_parameter_name]
        else:
            parameter_label = []

        if "beam_polarization" in keys:
            self.diffrn_radiation.polarization = ddict_diffrn["beam_polarization"]

        if "flipper_efficiency" in keys:
            self.diffrn_radiation.efficiency = ddict_diffrn["flipper_efficiency"]

        if "resolution_parameters" in keys:
            np_resolution = ddict_diffrn["resolution_parameters"]
            resolution = self.pd2d_instr_resolution
            resolution.u = float(np_resolution[0])
            resolution.v = float(np_resolution[1])
            resolution.w = float(np_resolution[2])
            resolution.x = float(np_resolution[3])
            resolution.y = float(np_resolution[4])

        if "resolution_phi_parameter" in keys:
            np_resolution = ddict_diffrn["resolution_phi_parameter"]
            resolution = self.pd2d_instr_resolution
            resolution.phi = float(np_resolution[0])

        if "background_intensity" in keys:
            background_intensity = ddict_diffrn["background_intensity"]
            self.pd2d_background.intensity = numpy.round(background_intensity, decimals=5)
            self.pd2d_background.form_gamma_nu_intensity()

        if "texture_g1" in keys:
            hh_g1 = ddict_diffrn["texture_g1"]
            hh_g2 = ddict_diffrn["texture_g2"]
            hh_axis = ddict_diffrn["texture_axis"]
            for i_item, item in enumerate(self.texture):
                item.g_1  = float(hh_g1[i_item])
                item.g_2  = float(hh_g2[i_item])
                item.h_ax = float(hh_axis[0, i_item])
                item.k_ax = float(hh_axis[1, i_item])
                item.l_ax = float(hh_axis[2, i_item])

        if "asymmetry_parameters" in keys:
            hh = ddict_diffrn["asymmetry_parameters"]
            if self.is_attribute("pd2d_instr_reflex_asymmetry"):
                asymmetry = self.pd2d_instr_reflex_asymmetry
            else:
                asymmetry = Pd2dInstrReflexAsymmetry()
                self.items.append(asymmetry)
            asymmetry.p1 = float(hh[0])
            asymmetry.p2 = float(hh[1])
            asymmetry.p3 = float(hh[2])
            asymmetry.p4 = float(hh[3])


        if "phase_scale" in keys:
            hh = ddict_diffrn["phase_scale"]
            for i_item, item in enumerate(self.phase.items):
                item.scale = float(hh[i_item])

        if "phase_ig" in keys:
            hh = ddict_diffrn["phase_ig"]
            for i_item, item in enumerate(self.phase.items):
                item.igsize = float(hh[i_item])

        if "offset_gamma" in keys:
            self.setup.offset_gamma = float(ddict_diffrn["offset_gamma"]) * 180./numpy.pi

        if "offset_nu" in keys:
            self.setup.offset_nu = float(ddict_diffrn["offset_nu"]) * 180./numpy.pi

        if "wavelength" in keys:
            self.setup.wavelength = float(ddict_diffrn["wavelength"])

        if len(parameter_label) > 0:
            for name, sigma in zip(l_parameter_name, l_sigma):
                if name[0] == "background_intensity":
                    self.pd2d_background.intensity_sigma[name[1]] = float(sigma)
                if name[0] == "phase_scale":
                    self.phase.items[name[1][0]].scale_sigma = float(sigma)
                if name[0] == "phase_ig":
                    self.phase.items[name[1][0]].igsize_sigma = float(sigma)
                if name[0] == "wavelength":
                    self.setup.wavelength_sigma = float(sigma)
                if name[0] == "offset_gamma":
                    self.setup.offset_gamma_sigma = float(sigma) * 180./numpy.pi
                if name[0] == "offset_nu":
                    self.setup.offset_nu_sigma = float(sigma) * 180./numpy.pi
                if name[0] == "beam_polarization":
                    self.diffrn_radiation.polarization_sigma = float(sigma)
                if name[0] == "flipper_efficiency":
                    self.diffrn_radiation.efficiency_sigma = float(sigma)
                if name[0] == "asymmetry_parameters":
                    asymmetry = self.pd2d_instr_reflex_asymmetry
                    if name[1][0] == 0:
                        asymmetry.p1_sigma = float(sigma)
                    elif name[1][0] == 1:
                        asymmetry.p2_sigma = float(sigma)
                    elif name[1][0] == 2:
                        asymmetry.p3_sigma = float(sigma)
                    elif name[1][0] == 3:
                        asymmetry.p4_sigma = float(sigma)
                if name[0] == "resolution_parameters":
                    resolution = self.pd2d_instr_resolution
                    if name[1][0] == 0:
                        resolution.u_sigma = float(sigma)
                    elif name[1][0] == 1:
                        resolution.v_sigma = float(sigma)
                    elif name[1][0] == 2:
                        resolution.w_sigma = float(sigma)
                    elif name[1][0] == 3:
                        resolution.x_sigma = float(sigma)
                    elif name[1][0] == 4:
                        resolution.y_sigma = float(sigma)
                if name[0] == "resolution_phi_parameter":
                    resolution = self.pd2d_instr_resolution
                    resolution.phi_sigma = float(sigma)
                if name[0] == "texture_g1":
                    self.texture.items[name[1][0]].g1_sigma = float(sigma)
                if name[0] == "texture_g2":
                    self.texture.items[name[1][0]].g2_sigma = float(sigma)
                if name[0] == "texture_axis":
                    ind_param, ind_a = name[1]
                    if ind_param == 0:
                        self.texture.items[ind_a].h_ax_sigma = float(sigma)
                    if ind_param == 1:
                        self.texture.items[ind_a].k_ax_sigma = float(sigma)
                    if ind_param == 2:
                        self.texture.items[ind_a].l_ax_sigma = float(sigma)

            if "background_intensity" in parameter_label:
                self.pd2d_background.form_gamma_nu_intensity()

        if (("signal_plus" in keys) and ("signal_minus" in keys) and ("signal_background" in keys)):
            pd2d_proc = Pd2dProc()
            pd2d_proc.gamma = numpy.round(ddict_diffrn["gamma"] * 180./numpy.pi, decimals=5)
            pd2d_proc.nu = numpy.round(ddict_diffrn["nu"] * 180./numpy.pi, decimals=5)
            pd2d_proc.intensity_plus_net = numpy.round(ddict_diffrn["signal_plus"], decimals=5)
            pd2d_proc.intensity_minus_net = numpy.round(ddict_diffrn["signal_minus"], decimals=5)
            pd2d_proc.intensity_bkg_calc = numpy.round(ddict_diffrn["signal_background"], decimals=5)
            pd2d_proc.excluded_points = ddict_diffrn["excluded_points"][:, :]
            pd2d_proc.form_gamma_nu_intensity_plus_net()
            pd2d_proc.form_gamma_nu_intensity_minus_net()
            pd2d_proc.form_gamma_nu_intensity_bkg_calc()
            pd2d_proc.form_gamma_nu_excluded_points()

            if "signal_exp_plus" in keys:
                pd2d_proc.intensity_plus = numpy.round(ddict_diffrn["signal_exp_plus"][0, :, :], decimals=5)
                pd2d_proc.intensity_plus_sigma = numpy.round(ddict_diffrn["signal_exp_plus"][1, :, :], decimals=5)
                pd2d_proc.intensity_minus = numpy.round(ddict_diffrn["signal_exp_minus"][0, :, :], decimals=5)
                pd2d_proc.intensity_minus_sigma = numpy.round(ddict_diffrn["signal_exp_minus"][1, :, :], decimals=5)
                pd2d_proc.form_gamma_nu_intensity_plus()
                pd2d_proc.form_gamma_nu_intensity_plus_sigma()
                pd2d_proc.form_gamma_nu_intensity_minus()
                pd2d_proc.form_gamma_nu_intensity_minus_sigma()
            elif "signal_exp" in keys:
                pd2d_proc.intensity = numpy.round(ddict_diffrn["signal_exp"][0, :, :], decimals=5)
                pd2d_proc.intensity_sigma = numpy.round(ddict_diffrn["signal_exp"][1, :, :], decimals=5)
                pd2d_proc.form_gamma_nu_intensity()
                pd2d_proc.form_gamma_nu_intensity_sigma()

            self.pd2d_proc = pd2d_proc

        l_pd_peak = []
        for item_phase in self.phase.items:
            s_label = f"dict_in_out_{item_phase.label:}"
            if s_label in keys:
                dict_crystal = ddict_diffrn[s_label]
                dict_crystal_keys = dict_crystal.keys()
                if (("index_hkl" in dict_crystal_keys) and  ("ttheta_hkl" in dict_crystal_keys) and  ("f_nucl" in dict_crystal_keys)):
                    pd_peak = PdPeakL(loop_name = item_phase.label)
                    int_max = numpy.square(numpy.abs(dict_crystal["f_nucl"]))

                    if "f_m_perp_o" in dict_crystal_keys:
                        int_max += numpy.sum(numpy.square(numpy.abs(dict_crystal["f_m_perp_o"])), axis=0)
                    if "p_1" in dict_crystal_keys:
                        int_max += dict_crystal["p_1"]

                    if "p_3" in dict_crystal_keys:
                        int_plus_max = int_max+dict_crystal["p_3"]
                        int_minus_max = int_max-dict_crystal["p_3"]
                    else:
                        int_plus_max = int_max
                        int_minus_max = int_max

                    index_hkl = dict_crystal["index_hkl"]
                    ttheta_hkl = dict_crystal["ttheta_hkl"]
                    lf = numpy.sin(ttheta_hkl)*numpy.sin(0.5*ttheta_hkl)
                    lf[lf<0] = 1.
                    pd_peak.numpy_index_h = index_hkl[0]
                    pd_peak.numpy_index_k = index_hkl[1]
                    pd_peak.numpy_index_l = index_hkl[2]
                    pd_peak.numpy_ttheta = numpy.round(ttheta_hkl * 180./numpy.pi, decimals=3)
                    pd_peak.numpy_intensity_plus = int_plus_max/lf
                    pd_peak.numpy_intensity_minus = int_minus_max/lf
                    if "multiplicity_hkl" in dict_crystal_keys:
                        pd_peak.numpy_index_mult = dict_crystal["multiplicity_hkl"]
                    if "ttheta_hkl" in dict_crystal_keys:
                        pd_peak.numpy_ttheta = numpy.round(dict_crystal["ttheta_hkl"] * 180./numpy.pi, decimals=3)
                    if (("gamma_hkl" in dict_crystal_keys) and ("nu_hkl" in dict_crystal_keys)):
                        gamma_hkl = dict_crystal["gamma_hkl"]
                        nu_hkl = dict_crystal["nu_hkl"]
                        pd_peak.numpy_gamma = numpy.round(gamma_hkl * 180./numpy.pi, decimals=3)
                        pd_peak.numpy_nu = numpy.round(nu_hkl * 180./numpy.pi, decimals=3)
                    pd_peak.numpy_to_items()
                    l_pd_peak.append(pd_peak)
        if len(l_pd_peak) > 0:
            self.add_items(l_pd_peak)

    def estimate_background(self):
        pd2d_background = self.pd2d_background
        intensity_refinement = numpy.copy(pd2d_background.intensity_refinement)
        pd2d_background.intensity_refinement = numpy.ones_like(intensity_refinement, dtype=bool)
        setup=self.setup
        pd2d_proc = self.pd2d_proc
        res = pd2d_proc.estimate_background(
            pd2d_background,
            offset_gamma=setup.offset_gamma,
            offset_nu = setup.offset_nu)
        pd2d_background.intensity_refinement = intensity_refinement
        pd2d_background.form_gamma_nu_intensity()
        return res



from typing import NoReturn
import re

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class AtomElectronConfiguration(ItemN):
    """
    Define the electron configuration for atoms.
    The elcetron configuration is separated on the core part which is spherical
    and non-spherical valence part. 

    The orientation and population of valence part is described in the object
    AtomRhoOrbitalValence of cryspy library.

    """
    ATTR_MANDATORY_NAMES = ("label", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = ("core", "valence")
    ATTR_OPTIONAL_TYPES = (str, str)
    ATTR_OPTIONAL_CIF = ("core", "valence")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_electron_configuration"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomElectronConfiguration, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"mosaicity": 0., "radius": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def get_core_shells_populations(self):
        """
        Give list of shels and their population for core part
        """
        l_res = []
        s_core = self.core
        if s_core is not None:
            for _s_shell in s_core.split():
                m = re.match(r"(\d*)([s,p,d,f])(\d*)", _s_shell)
                if m is not None:
                    res = (f"{m.group(1):}{m.group(2):}", int(m.group(3)))
                    l_res.append(res)
        return l_res

    def get_valence_shells(self):
        """
        Give list of shels for valence part
        """
        l_res = []
        s_valence = self.valence
        if s_valence is not None:
            for _s_shell in s_valence.split():
                m = re.match(r"(\d*)([s,p,d,f])", _s_shell)
                if m is not None:
                    res = f"{m.group(1):}{m.group(2):}"
                    l_res.append(res)
        return l_res

class AtomElectronConfigurationL(LoopN):
    """
    Define the electron configuration for atoms.
    The elcetron configuration is separated on the core part which is spherical
    and non-spherical valence part. 

    The orientation and population of valence part is described in the object
    AtomRhoOrbitalValence of cryspy library.
    """
    ITEM_CLASS = AtomElectronConfiguration
    ATTR_INDEX = "label"
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomElectronConfigurationL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name
        
    def get_core_shells_populations(self):
        """
        Give list of shels and their population for core part of all atoms 
        """
        ll_res = [item.get_core_shells_populations() for item in self.items]
        return ll_res

# s_cont = """
# loop_
# _atom_electron_configuration_label
# _atom_electron_configuration_core
# _atom_electron_configuration_valence
#  O1 "1s2" "2s 2p"
#  O2 "1s2" "2s 2p"
#  O3 "1s2" "2s 2p"
# Co1 "1s2 2s2 2p6 3s2 3p6" .
# Co2 "1s2 2s2 2p6 3s2 3p6" .
#  V1 "1s2 2s2 2p6 3s2 3p6" .
# """

# obj = AtomElectronConfigurationL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj["V1"].get_core_shells_populations(), end="\n\n")
